unit NkVImage;

interface

uses Windows,
  SUtil, SList, npi, in_dib, in_spi;

type
  {
  TSDib = record
    Bits: Pointer;
    case Integer of
      1: (Header: TBitmapInfoHeader);
      2: (Info: TBitmapInfo);
      3: (Dummy: array[0..SizeOf(TBitmapInfoHeader)+SizeOf(TRGBQuad)*259] of Byte);
  end;
  }

  PMaxBitmapInfo = ^TMaxBitmapInfo;
  TMaxBitmapInfo = packed record
    bmiHeader: TBitmapInfoHeader;
    bmiColors: array[0..255] of TRGBQuad;
  end;

  TFilterList = class(TObject)
  private
    FFileNames: TSStrList;
    FFormats: TSStrList;
  public
    DibFilter: TDibFilter;
    SpiFilter: TSpiFilter;
    constructor Create;
    destructor Destroy; override;
    function LookUp(InBuf: PChar; InBufSize: Integer): TImageFilter;
    procedure Add(const FileName: string; const FormatName: string);
    property FileNames: TSStrList read FFileNames;
    property Formats: TSStrList read FFormats;
  end;

  TNkVImage = class(TObject)
  private
    SPIHBm: HLOCAL;
    function Decode(const Filter: TImageFilter; InBuf: PChar; InBufSize: Integer): Integer;
  public
    FilterList: TFilterList;
    Loaded: Boolean;
    Width: Integer;
    Height: Integer;
    FileSize: Integer;
    MemSize: Integer;
    BitPerPixelOrg: Integer;
    FormatName: string;
    FilterName: string;
    FilterFileName: string;
    Memo: string;
    BitmapInfo: TMaxBitmapInfo;
    Bits: Pointer;
    hAMTmp: THandle;
    DecodeResult: Integer;
    ErrFilterFileName: string;
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function LoadFromMemory(InBuf: Pointer; InBufSize: Integer): Integer;
    function LoadFromFile(const FileName: string): Integer;
    function SaveToFile(const FileName: string): Integer;
  end;

const
  nilfFilterNotFound = -1;
  nilfArchive = -2;
  nilfDecodeErr = -3;
  TestBufSize = 2048; // tH[}bĝ߂ɓǂݍbyte

implementation

procedure GetImageFormat(pBuffer: PChar; var Name: string; var FullName: string);
begin
  Name := '';
  FullName := '';
  if Copy(pBuffer, 1, 3) = 'GIF' then
  begin
    Name := 'gif';
    FullName := Copy(pBuffer, 1, 6);
  end
  else if Copy(pBuffer, 1, 2) = #$FF#$D8 then
  begin
    Inc(pBuffer, 6);
    if Copy(pBuffer, 1, 4) = 'JFIF' then
    begin
      Name := 'jpeg';
      FullName := 'JPEG - JFIF'
    end
    else if Copy(pBuffer, 1, 4) = 'Exif' then
    begin
      Name := 'jpeg';
      FullName := 'Exif';
    end;
  end
  else if Copy(pBuffer, 1, 8) = #$89'PNG'#$0D#$0A#$1A#$0A then
  begin
    Name := 'png';
    FullName := 'PNG';
    case pBuffer[28] of
      #0: FullName := FullName + ' - non-interlaced';
      #1: FullName := FullName + ' - interlaced';
    end;
    if Ord(pBuffer[25]) and 4 <> 0 then
      FullName := FullName + ' (with alpha)';
  end
  else if (pBuffer[0] = 'P') and (pBuffer[1] in ['1'..'7']) and (pBuffer[2] in [' ', '#', #9, #13, #10]) then
  begin
    Name := 'pnm';
    case pBuffer[1] of
      '1': FullName := 'PBM image text';
      '2': FullName := 'PGM image text';
      '3': FullName := 'PPM image text';
      '4': FullName := 'PBM "rawbits" image data';
      '5': FullName := 'PGM "rawbits" image data';
      '6': FullName := 'PPM "rawbits" image data';
      '7': FullName := 'PAM image file';
    end;
  end;
end;

{ TFilterList }

constructor TFilterList.Create;
begin
  FFileNames := TSStrList.Create;
  FFormats := TSStrList.Create;
  DibFilter := TDibFilter.Create;
  SpiFilter := TSPIFilter.Create;
end;

destructor TFilterList.Destroy;
begin
  SpiFilter.Free;
  DibFilter.Free;
  FFormats.Free;
  FFileNames.Free;
  inherited;
end;

procedure TFilterList.Add(const FileName: string; const FormatName: string);
begin
  FFileNames.Add(FileName);
  FFormats.Add(FormatName);
end;

function TFilterList.LookUp(InBuf: PChar; InBufSize: Integer): TImageFilter;
var
  Buf: array[0..TestBufSize-1] of Char;
  i: Integer;
  Fmt, dummy: string;
begin
  // tB^obt@̒gj󂷂邩Ȃ̂ŃRs[n
  FillChar(Buf, SizeOf(Buf), 0);
  if InBufSize < SizeOf(Buf) then
    Move(InBuf^, Buf, InBufSize)
  else
    Move(InBuf^, Buf, SizeOf(Buf));

  if DibFilter.Test(Buf, SizeOf(Buf)) then
  begin
    Result := DibFilter;
    Exit;
  end;

  // FormatŌ
  GetImageFormat(Buf, Fmt, dummy);
  for i := 0 to FFormats.Count-1 do
  begin
    if Fmt = FFormats[i] then
    begin
      if SpiFilter.Load(FFileNames[i]) then
      begin
        if SpiFilter.Test(Buf, SizeOf(Buf)) then
        begin
          Result := SpiFilter;
          Exit;
        end;
        SpiFilter.UnLoad;
      end;
    end;
  end;

  // vOCɔ肳
  for i := 0 to FFileNames.Count-1 do
  begin
    if SpiFilter.Load(FFileNames[i]) then
    begin
      if SpiFilter.Test(Buf, SizeOf(Buf)) then
      begin
        Result := SpiFilter;
        Exit;
      end;
      SpiFilter.UnLoad;
    end;
  end;

  Result := nil;
end;

{ TNkVImage }

constructor TNkVImage.Create;
begin
  FilterList := TFilterList.Create;
end;

destructor TNkVImage.Destroy;
begin
  Clear;
  FilterList.Free;
end;

procedure TNkVImage.Clear;
begin
  if SPIHBm = 0 then
  begin
    if Bits <> nil then
      GlobalFree(HGLOBAL(Bits))
  end else
  begin
    LocalUnLock(SPIHBm);
    LocalFree(SPIHBm);
    SPIHBm := 0;
  end;
  Bits := nil;
  Loaded := False;
  Memo := '';
  FilterFileName := '';
end;

function TNkVImage.Decode(const Filter: TImageFilter; InBuf: PChar; InBufSize: Integer): Integer;
var
  BitmapInfo: TMaxBitmapInfo;
  pDst: Pointer;
  DstSize: Integer;
  Fmt, Comment: string;
  HBm: HLOCAL;

  ImageInfo: TImageInfo;
  ImageMetaInfo: TImageMetaInfo;
  TmpFilterName: string;
begin
  FillChar(ImageInfo, SizeOf(ImageInfo), 0);

  Filter.GetInfo(InBuf, InBufSize, ImageInfo, @ImageMetaInfo);
  TmpFilterName := Filter.Name;

  if Filter is TDibFilter then
  begin
    HBm := 0;
    DstSize := ((ImageInfo.Width * ImageInfo.BitPerPixel + 31) div 32) * 4 * ImageInfo.Height;
    pDst := Pointer(GlobalAlloc(0, DstSize));
    // todo: ԂlƎg
    Result := (Filter as TDibFilter).Decode2(InBuf, InBufSize, pDst, DstSize,
      PBitmapInfo(@BitmapInfo));
  end
  else begin
    // SPI Plug-in
    Result := (Filter as TSpiFilter).Decode2(InBuf, InBufSize, HBm, PBitmapInfo(@BitmapInfo));
    pDst := LocalLock(HBm);
    DstSize := LocalSize(HBm);
    Comment := ConvLineBreak(Copy(ImageMetaInfo.Comment, 1, Length(ImageMetaInfo.Comment)), #13#10);
    Filter.FreeImageMetaInfo(ImageMetaInfo);
  end;

  if Result <> 0 then
  begin
    if Filter is TSpiFilter then
    begin
      ErrFilterFileName := (Filter as TSpiFilter).FileName;
      (Filter as TSpiFilter).UnLoad;
    end;
    Exit;
  end;

  Clear;
  Loaded := True;
  SPIHBm := HBm;
  MemSize := DstSize;
  Self.BitmapInfo := BitmapInfo;
  Width := ImageInfo.Width;
  Height := ImageInfo.Height;
  BitPerPixelOrg := ImageMetaInfo.SrcBitPerPixel;
  Bits := pDst;
  Memo := Comment;
  FormatName := ImageMetaInfo.FormatName;
  if FormatName = '' then
    GetImageFormat(InBuf, Fmt, FormatName);
  FilterName := TmpFilterName;
  FileSize := InBufSize;
  if Filter is TSpiFilter then
  begin
    FilterFileName := (Filter as TSpiFilter).FileName;
    (Filter as TSpiFilter).UnLoad;
  end;
end;

function TNkVImage.LoadFromMemory(InBuf: Pointer; InBufSize: Integer): Integer;
var
  Filter: TImageFilter;
begin
  Filter := FilterList.Lookup(InBuf, InBufSize);
  if not Assigned(Filter) then
  begin
    Result := nilfFilterNotFound;
    Exit;
  end;

  if SafeIsArchivePlugin(Filter) then
  begin
    hAMTmp := (Filter as TSpiFilter).Handle;
    Result := nilfArchive;
  end
  else begin
    DecodeResult := Decode(Filter, InBuf, InBufSize);
    if DecodeResult = 0 then
      Result := 0
    else
      Result := nilfDecodeErr;
  end;
end;

function TNkVImage.LoadFromFile(const FileName: string): Integer;
const
  StdinBufSize = 32768;
var
  InBuf: PChar;
  InBufSize: Integer;
  ReadSize: Integer;
  HeadBuf: array[0..TestBufSize-1] of Char;
  Filter: TImageFilter;
  z: integer;
  F: TSFile;
begin
  hAMTmp := 0;

  F := TSFile.Create(FileName, 'r');
  try
    Result := F.ResultCode;
    if Result <> 0 then
      Exit;

    FillChar(HeadBuf, SizeOf(HeadBuf), 0);
    ReadSize := F.Read(HeadBuf, SizeOf(HeadBuf));
    Result := F.ResultCode;
    if Result <> 0 then
      Exit;

    Filter := FilterList.Lookup(HeadBuf, ReadSize);
    if not Assigned(Filter) then
    begin
      Result := nilfFilterNotFound;
      Exit;
    end;

    if SafeIsArchivePlugin(Filter) then
    begin
      hAMTmp := (Filter as TSpiFilter).Handle;
      Result := nilfArchive;
      Exit;
    end;

    if FileName = '' then
    begin
      // read from STDIN
      InBufSize := ReadSize;
      GetMem(InBuf, InBufSize);
      Move(HeadBuf, InBuf^, InBufSize);
      z := 2;
      while z > 0 do // 2ǂœǂݎoCg0ȂIĂƂ炵
      begin
        ReallocMem(InBuf, InBufSize+StdinBufSize);
        ReadSize := F.Read(InBuf[InBufSize], StdinBufSize);
        if (F.ResultCode <> 0) or (ReadSize = 0) then
          Dec(z)
        else
          Inc(InBufSize, ReadSize);
      end;
    end else
    begin
      InBufSize := F.Size;
      GetMem(InBuf, InBufSize);
      Move(HeadBuf, InBuf^, ReadSize);
      F.Read(InBuf[ReadSize], InBufSize-ReadSize);
      Result := F.ResultCode;
      if Result <> 0 then
      begin
        FreeMem(InBuf);
        Exit;
      end;
    end;
  finally
    F.Free;
  end;
  DecodeResult := Decode(Filter, InBuf, InBufSize);
  if DecodeResult <> 0 then
    Result := nilfDecodeErr;
  FreeMem(InBuf);
end;

function TNkVImage.SaveToFile(const FileName: string): Integer;
var
  bfh: TBitmapFileHeader;
  bmisize: Integer;
begin
  with TSFile.Create(FileName, 'w') do
    try
      Result := ResultCode;
      if Result <> 0 then
        Exit;

      with BitmapInfo do
      begin
        bmisize := SizeOf(bmiHeader);
        if bmiHeader.biBitCount <= 8 then
          Inc(bmisize, (1 shl bmiHeader.biBitCount) * SizeOf(TRGBQUAD));
      end;

      FillChar(bfh, SizeOf(bfh), 0);
      bfh.bfType := Byte('B') + Byte('M') shl 8;
      bfh.bfSize := SizeOf(bfh) + bmisize + MemSize;
      bfh.bfOffBits := SizeOf(bfh) + bmisize;
      // vG[`FbN
      Write(bfh, SizeOf(bfh));
      Write(BitmapInfo, bmisize);
      Write(Bits^, MemSize);
    finally
      Free;
    end;
end;

end.
