unit ResizeImage;

interface

uses Windows;

procedure ResizeDIB(DestBmi: PBitmapInfo; var DestBits: Pointer;
  DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer;
  const SrcBmi: TBitmapInfo; SrcBits: Pointer; Filter: Boolean);

implementation

//uses SUtil;

type
  PByteArray = ^TByteArray;
  TByteArray = array[0..MaxInt-1] of Byte;
  PIntArray = ^TIntArray;
  TIntArray = array[0..(MaxInt div SizeOf(Integer))-1] of Integer;
  PPointerArray = ^TPointerArray;
  TPointerArray = array[0..MaxInt div SizeOf(Integer)-1] of Pointer;

{
var
  SqrtTable: array[0..65535] of Byte;

procedure MakeSqrtTable;
var
  i: Integer;
begin
  for i := 0 to High(SqrtTable) do
    SqrtTable[i] := Trunc(Sqrt(i));
end;
}

function CalcPitch(Width: Integer; Bpp: Integer): Integer;
begin
  Result := (Width * Bpp shr 3 + 3) and not 3;
end;

procedure Nearest(SrcBits: Pointer; SrcWidth, SrcHeight: Integer;
  DestBits: Pointer; DestWidth, DestHeight: Integer;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer; Bpp: Integer);
var
  zx, zy: Double;
  sx: PIntArray;
  sy: PPointerArray;
  x, y: Integer;
  sp: PByteArray;
  dp, dsp: PByte;
  c0, c1: Byte;
  SrcPitch, ViewPitch: Integer;
begin
  SrcPitch := CalcPitch(SrcWidth, Bpp);
  ViewPitch := CalcPitch(ViewWidth, Bpp);

  zx := SrcWidth / DestWidth;
  zy := SrcHeight / DestHeight;

  // 摜̑ΉsNZ߂
  GetMem(sx, SizeOf(Integer) * ViewWidth);
  for x := 0 to ViewWidth-1 do
    sx[x] := Trunc(zx * (ViewLeft+x+0.5));
  GetMem(sy, SizeOf(Pointer) * ViewHeight);
  for y := 0 to ViewHeight-1 do
    sy[y] := @PByteArray(SrcBits)[Trunc(zy * (ViewTop+y+0.5)) * SrcPitch];
  dsp := DestBits;

  case Bpp of
    1:
      begin
        FillChar(DestBits^, ViewPitch * ViewHeight, 0);
        for y := 0 to ViewHeight-1 do
        begin
          dp := dsp;
          sp := sy[y];
          x := 0;
          while x < ViewWidth do
          begin
            PByteArray(dp)[x div 8] := PByteArray(dp)[x div 8] or
              (sp[sx[x] div 8] shl (sx[x] mod 8) and $80) shr (x mod 8);
            Inc(x);
          end;
          Inc(dsp, ViewPitch);
        end;
      end;
    4:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dsp;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          if sx[x] mod 2 = 0 then
            c0 := (sp[sx[x] div 2] and $F0) shr 4
          else
            c0 := sp[sx[x] div 2] and $F;
          if sx[x+1] mod 2 = 0 then
            c1 := (sp[sx[x+1] div 2] and $F0) shr 4
          else
            c1 := sp[sx[x+1] div 2] and $F;
          dp^ := c0 shl 4 + c1;
          Inc(dp);
          Inc(x, 2);
        end;
        Inc(dsp, ViewPitch);
      end;
    8:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dsp;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          dp^ := sp[sx[x]];
          Inc(dp);
          Inc(x);
        end;
        Inc(dsp, ViewPitch);
      end;
    16:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dsp;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          PWord(dp)^ := PWord(@sp[sx[x]*2])^;
          Inc(dp, 2);
          Inc(x);
        end;
        Inc(dsp, ViewPitch);
      end;
    24:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dsp;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          PInteger(dp)^ := PInteger(@sp[sx[x]*3])^;
          Inc(dp, 3);
          Inc(x);
        end;
        Inc(dsp, ViewPitch);
      end;
    32:
      for y := 0 to ViewHeight-1 do
      begin
        dp := dsp;
        sp := sy[y];
        x := 0;
        while x < ViewWidth do
        begin
          PInteger(dp)^ := PInteger(@sp[sx[x]*4])^;
          Inc(dp, 4);
          Inc(x);
        end;
        Inc(dsp, ViewPitch);
      end;
  end;

  FreeMem(sx);
  FreeMem(sy);
end;

procedure Bilinear24(SrcBits: Pointer; SrcWidth, SrcHeight: Integer;
  DestBits: Pointer; DestWidth, DestHeight: Integer;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
var
  zx, zy, t, tx, ty: Double;
  sx, sy: PIntArray;
  x, y, nextline: Integer;
  px, py: Byte;
  dp, dsp: PByte;
  sp, sp0, sp1, sp2, sp3: PByteArray;
  f0, f1, f2, f3: Integer;
  SrcPitch, ViewPitch: Integer;
  c, i: Integer;
begin
  // 24bpp only
  SrcPitch := CalcPitch(SrcWidth, 24);
  ViewPitch := CalcPitch(ViewWidth, 24);

  zx := SrcWidth / DestWidth;
  zy := SrcHeight / DestHeight;

  // ␳l
  tx := 0.5-(1/zx)/2.0;
  ty := 0.5-(1/zy)/2.0;

  // 摜̑ΉsNZ߂
  GetMem(sx, SizeOf(Integer) * ViewWidth);
  for x := 0 to ViewWidth-1 do
  begin
    t := zx * (ViewLeft+x+tx);
    if t < 0 then
      t := 0;
    sx[x] := Trunc(t * $10000);
  end;
  GetMem(sy, SizeOf(Integer) * ViewHeight);
  for y := 0 to ViewHeight-1 do
  begin
    t := zy * (ViewTop+y+ty);
    if t < 0 then
      t := 0;
    sy[y] := Trunc(t * $10000);
  end;

  //c := GetTickCount; for i := 0 to 9 do begin    //312

  dsp := DestBits;
  for y := 0 to ViewHeight-1 do
  begin
    dp := dsp;
    sp := PByteArray(@PByteArray(SrcBits)[(sy[y] shr 16)*SrcPitch]);
    nextline := Ord((sy[y] shr 16)<SrcHeight-1)*SrcPitch;
    //  +---+---+
    //  |sp2|sp3|  py
    //  +---+---+  ^
    //  |sp0|sp1|  |
    //  +---+---+  +->px
    // sp0*(1-px)*(1-py) + sp1*px*(1-py) + sp2*(1-px)*py + sp3*px*py;
    py := sy[y] shr 8;
    x := 0;
    while x < ViewWidth do // forwhilêقc
    begin
      px := sx[x] shr 8; //ɂȂƒxȂ

      // |C^ẑق
      sp0 := PByteArray(@sp[(sx[x] shr 16)*3]);
      sp1 := PByteArray(@sp0[Ord((sx[x] shr 16)<SrcWidth-1)*3]);
      sp2 := PByteArray(@sp0[nextline]);
      sp3 := PByteArray(@sp1[nextline]);

      f0 := ($100-px) * ($100-py);
      f1 := px * ($100-py);
      f2 := ($100-px) * py;
      f3 := px * py;

      dp^ := (sp0[0]*f0 + sp1[0]*f1 + sp2[0]*f2 + sp3[0]*f3) shr 16;
      Inc(dp);
      dp^ := (sp0[1]*f0 + sp1[1]*f1 + sp2[1]*f2 + sp3[1]*f3) shr 16;
      Inc(dp);
      dp^ := (sp0[2]*f0 + sp1[2]*f1 + sp2[2]*f2 + sp3[2]*f3) shr 16;
      Inc(dp);

      Inc(x);
    end;
    Inc(dsp, ViewPitch);
  end;

  //end; MessageBox(0, pchar(IStr(GetTickCount-c)), nil, 0);

  FreeMem(sx);
  FreeMem(sy);
end;

procedure Average24(SrcBits: Pointer; SrcWidth, SrcHeight: Integer;
  DestBits: Pointer; DestWidth, DestHeight: Integer;
  ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer);
var
  SrcPitch, ViewPitch: Integer;
  zx, zy: Double;
  sx, sy, fx, fy: PIntArray;
  x, y, yy, nx, ny: Integer;
  px, py, px2, py2: Word;
  sp0, dp, dsp: PByte;
  ssp, sp: PByteArray;
  c, i, j, k, ssx, ssy: Integer;
  dr, dg, db, fb ,fg, fr: Integer;
  yb, ybuf: PIntArray;
  lastline: Boolean;
begin
  // 24bpp only
  SrcPitch := CalcPitch(SrcWidth, 24);
  ViewPitch := CalcPitch(ViewWidth, 24);

  zx := SrcWidth / DestWidth;
  zy := SrcHeight / DestHeight;
  nx := Trunc($10000 / zx);
  ny := Trunc($10000 / zy);

  // 摜̑ΉsNZ߂
  // ݂̃sNZƎ̃sNZ̍KvȂ̂Ŏۂ̃TCY+1܂ŕKv
  GetMem(sx, SizeOf(Integer) * ViewWidth + 1);
  GetMem(fx, SizeOf(Integer) * ViewWidth + 1);
  for x := 0 to ViewWidth do
  begin
    sx[x] := Trunc((zx*(ViewLeft+x)) * $10000) shr 16;
    fx[x] := Trunc((zx*(ViewLeft+x)) * $10000) and $FFFF;
  end;
  GetMem(sy, SizeOf(Integer) * ViewHeight + 1);
  GetMem(fy, SizeOf(Integer) * ViewHeight + 1);
  for y := 0 to ViewHeight do
  begin
    sy[y] := Trunc((zy*(ViewTop+y)) * $10000) shr 16;
    fy[y] := Trunc((zy*(ViewTop+y)) * $10000) and $FFFF;
  end;

  GetMem(ybuf, (Trunc(zy)+2)*SizeOf(Integer)*3);  // xɕςsNZ̃obt@

  //c := GetTickCount; for k := 0 to 9 do begin     //266

  dsp := DestBits;
  for y := 0 to ViewHeight-1 do
  begin
    dp := dsp;
    sp := PByteArray(@PByteArray(SrcBits)[sy[y]*SrcPitch]);
    yy := sy[y+1]-sy[y];
    py := $10000 - fy[y];
    py2 := fy[y+1];
    ssy := yy-Ord(py<>0)-1;
    lastline := ViewTop+y = DestHeight-1;

    x := 0;
    while x < ViewWidth do
    begin
      yb := ybuf;
      px := $10000 - fx[x];
      px2 := fx[x+1];
      ssx := sx[x+1]-sx[x]-1-Ord(px<>0);
      for i := 0 to yy do
      begin
        if lastline then
          ssp := PByteArray(@sp[SrcPitch*(i-1)])
        else
          ssp := PByteArray(@sp[SrcPitch*i]);
        sp0 := @ssp[sx[x]*3];

        dr := 0;
        dg := 0;
        db := 0;
        fr := 0;
        fg := 0;
        fb := 0;

        if px <> $0000 then
        begin
          Inc(fb, sp0^ * px);
          Inc(sp0);
          Inc(fg, sp0^ * px);
          Inc(sp0);
          Inc(fr, sp0^ * px);
          Inc(sp0);
        end;
        for j := 0 to ssx do
        begin
          Inc(db, sp0^);
          Inc(sp0);
          Inc(dg, sp0^);
          Inc(sp0);
          Inc(dr, sp0^);
          Inc(sp0);
        end;
        if px2 <> 0 then
        begin
          Inc(fb, sp0^ * px2);
          Inc(sp0);
          Inc(fg, sp0^ * px2);
          Inc(sp0);
          Inc(fr, sp0^ * px2);
        end;

        yb[0] := ((db + fb shr 16) * nx + $8000) shr 16;
        yb[1] := ((dg + fg shr 16) * nx + $8000) shr 16;
        yb[2] := ((dr + fr shr 16) * nx + $8000) shr 16;
        yb := PIntArray(@yb[3]);
      end;

      dr := 0;
      dg := 0;
      db := 0;
      fr := 0;
      fg := 0;
      fb := 0;

      yb := ybuf;
      if py <> $0000 then
      begin
        Inc(fb, yb[0] * py);
        Inc(fg, yb[1] * py);
        Inc(fr, yb[2] * py);
        yb := PIntArray(@yb[3]);
      end;
      for j := 0 to ssy do
      begin
        Inc(db, yb[0]);
        Inc(dg, yb[1]);
        Inc(dr, yb[2]);
        yb := PIntArray(@yb[3]);
      end;
      if py2 <> 0 then
      begin
        Inc(fb, yb[0] * py2);
        Inc(fg, yb[1] * py2);
        Inc(fr, yb[2] * py2);
      end;

      // kƌ덷̒~ςŐFÂȂ̂$8000𑫂
      dp^ := ((db + fb shr 16) * ny + $8000) shr 16;
      Inc(dp);
      dp^ := ((dg + fg shr 16) * ny + $8000) shr 16;
      //if dp^ = 254 then dp^ := 0;
      Inc(dp);
      dp^ := ((dr + fr shr 16) * ny + $8000) shr 16;
      //if dp^ = 253 then dp^ := 0;
      Inc(dp);

      Inc(x);
    end;
    Inc(dsp, ViewPitch);
  end;

  //end; MessageBox(0, pchar(IStr(GetTickCount-c)), nil, 0);

  FreeMem(sx);
  FreeMem(sy);
  FreeMem(fx);
  FreeMem(fy);
  FreeMem(ybuf);
end;

(*
// ʐϕϖ@1/ɏkĂoCjAŖړĨTCYɂ
procedure Average24b(SrcBits: Pointer; SrcWidth, SrcHeight: Integer;
  DestBits: Pointer; DestWidth, DestHeight: Integer;
  ViewLeft, ViewTop, ViewRight, ViewBottom: Integer);
var
  zx, zy: Integer;
  x, y: Integer;
  sp, dp, ssp, p: PByte;
  SrcPitch, ViewPitch: Integer;
  n, c, k, i, j, dr, dg, db: Integer;
  SrcWidth2, SrcHeight2, SrcPitch2: Integer;
  SrcBits2: Pointer;
begin
  // 24bpp only
  SrcPitch := CalcPitch(SrcWidth, 24);
  ViewPitch := CalcPitch(ViewRight-ViewLeft, 24);

  zx := SrcWidth div DestWidth;
  zy := SrcHeight div DestHeight;
  {
  if ((SrcWidth mod DestWidth) <> 0) and (zx > 2) then
    zx := zx - 1;
  if ((SrcHeight mod DestHeight) <> 0) and (zy > 2) then
    zy := zy - 1;
  }
  n := $10000 div (zx*zy);
  SrcWidth2 := SrcWidth div zx;
  SrcHeight2 := SrcHeight div zy;
  SrcPitch2 := CalcPitch(SrcWidth2, 24);
  GetMem(SrcBits2, SrcPitch2*SrcHeight2);

  // 156
  //c := GetTickCount; for k := 0 to 9 do
  for y := 0 to SrcHeight2-1 do
  begin
    dp := @PByteArray(SrcBits2)[y*SrcPitch2];
    sp := @PByteArray(SrcBits)[y*zy*SrcPitch];
    x := 0;
    while x < SrcWidth2 do
    begin
      dr := 0;
      dg := 0;
      db := 0;
      ssp := sp;
      for i := 0 to zy-1 do
      begin
        p := ssp;
        for j := 0 to zx-1 do
        begin
          {
          Inc(dr, PByteArray(p)[0]);
          Inc(dg, PByteArray(p)[1]);
          Inc(db, PByteArray(p)[2]);
          }
          Inc(db, PByteArray(p)[0]*PByteArray(p)[0]);
          Inc(dg, PByteArray(p)[1]*PByteArray(p)[1]);
          Inc(dr, PByteArray(p)[2]*PByteArray(p)[2]);
          Inc(p, 3);
        end;
        Inc(ssp, SrcPitch);
      end;
      {
      dp^ := dr * n shr 16;
      Inc(dp);
      dp^ := dg * n shr 16;
      Inc(dp);
      dp^ := db * n shr 16;
      Inc(dp);
      }

      // ganmma 2.0̂
      dp^ := SqrtTable[db * n shr 16];
      Inc(dp);
      dp^ := SqrtTable[dg * n shr 16];
      Inc(dp);
      dp^ := SqrtTable[dr * n shr 16];
      Inc(dp);

      Inc(sp, 3*zx);
      Inc(x);
    end;
  end;

  if (SrcWidth2 = DestWidth) and (SrcHeight2 = DestHeight) then
  begin
    for y := ViewTop to ViewBottom-1 do
    begin
      dp := @PByteArray(DestBits)[(y-ViewTop)*ViewPitch];
      sp := @PByteArray(SrcBits2)[y*SrcPitch2+ViewLeft*3];
      Move(sp^, dp^, ViewPitch);
    end;
  end else
    Bilinear24(SrcBits2, SrcWidth2, SrcHeight2, DestBits, DestWidth, DestHeight,
      ViewLeft, ViewTop, ViewRight, ViewBottom);

  //MessageBox(0, pchar(SStr(GetTickCount-c)), nil, 0);
  FreeMem(SrcBits2);
end;
*)

procedure ResizeDIB(DestBmi: PBitmapInfo; var DestBits: Pointer;
  DestWidth, DestHeight, ViewLeft, ViewTop, ViewWidth, ViewHeight: Integer;
  const SrcBmi: TBitmapInfo; SrcBits: Pointer; Filter: Boolean);
var
  SrcWidth, SrcHeight: Integer;
  ViewPitch: Integer;
begin
  if DestHeight < ViewHeight then
  begin
    ViewTop := 0;
    ViewHeight := DestHeight;
  end else
    ViewTop := DestHeight - ViewTop - ViewHeight; // DIBȂ̂ŃCi[t
  if DestWidth < ViewWidth then
  begin
    ViewLeft := 0;
    ViewWidth := DestWidth;
  end;
  ViewPitch := CalcPitch(ViewWidth, SrcBmi.bmiHeader.biBitCount);

  SrcWidth := SrcBmi.bmiHeader.biWidth;
  SrcHeight := SrcBmi.bmiHeader.biHeight;

  //DelphĩǗ}l[WXy[XfꂽubN
  //ԂꍇSetDIBitɎs\̂GetMemł͂ȂGlobalLockg
  // +1͊gk݂̂͂
  DestBits := Pointer(GlobalAlloc(GMEM_FIXED, ViewPitch * (ViewHeight+1)));

  if Filter and (SrcBmi.bmiHeader.biBitCount = 24) then
  begin
    if (DestWidth > SrcWidth) or (SrcWidth div DestWidth < 2) then
      Bilinear24(SrcBits, SrcWidth, SrcHeight, DestBits, DestWidth, DestHeight,
        ViewLeft, ViewTop, ViewWidth, ViewHeight)
    else
      Average24(SrcBits, SrcWidth, SrcHeight, DestBits, DestWidth, DestHeight,
        ViewLeft, ViewTop, ViewWidth, ViewHeight);
  end else
    Nearest(SrcBits, SrcWidth, SrcHeight, DestBits, DestWidth, DestHeight,
      ViewLeft, ViewTop, ViewWidth, ViewHeight, SrcBmi.bmiHeader.biBitCount);

  Move(SrcBmi, DestBmi^, SizeOf(TBitmapInfoHeader)+SizeOf(TRGBQuad)*256);
  DestBmi.bmiHeader.biWidth := ViewWidth;
  DestBmi.bmiHeader.biHeight := ViewHeight;
end;

//initialization
//  MakeSqrtTable;

end.
