unit SCmd;

interface

uses Windows, SConfFile;

const
  IDM_SCMD_FIRST = 20000; // ̒lIDU

type
  PCmd = ^TCmd;
  TCmd = record
    Name: string;
    KeyName: string;
    Id: Integer;
  end;

  PCmdTable = ^TCmdTable;
  TCmdTable = array[0..0] of TCmd;

  TKeyState = set of (ksAlt, ksCtrl, ksShift, ksWin);

  TSCmd = class(TObject)
  private
    FKeyList: array[0..255] of TAccel;
    FKeyListCount: Integer;
    FCmdTable: PCmdTable;
    FCmdTableCount: Integer;
    function StrToAccel(const KeyName: string; const ActionName: string): TAccel;
    function AccelToStr(const Accel: TAccel): string;
    procedure LoadKeymapFromFile(const FileName: string);
    procedure LoadKeymapFromArray(const Keymap: array of string);
    function NameToCmd(const Name: string): PCmd;
    function IdToCmd(Id: Integer): PCmd;
  public
    AccHandle: HACCEL;
    constructor Create(const CmdList: array of string; const DefaultKeys: array of string; FileName: string);
    destructor Destroy; override;
    function Id(const Name: string): Integer;
    function Name(Id: Integer): string;
  end;

  TCmdMenuItem = record
    Caption: string;
    CmdName: string;
    Enabled: Boolean;
    Checked: Boolean;
  end;
  TCmdMenuItemArray = array[0..0] of TCmdMenuItem;
  PCmdMenuItemArray = ^TCmdMenuItemArray;

  TSCmdMenu = class(TObject)
  private
    FItems: PCmdMenuItemArray;
    FItemsCount: Integer;
    FCmd: TSCmd;
  public
    constructor Create(Cmd: TSCmd; const MenuItemTable: array of string);
    destructor Destroy; override;
    function MakeMenuHandle: HMENU;
    procedure DisableItems(const Names: array of string);
    procedure CheckItem(const Name: string; Checked: Boolean);
  end;

procedure StrToKey(const KeyName: string; var KeyState: TKeyState; var VKey: UINT);

implementation

type
 TSpKeyTable = record
   Name: string;
   Id: Integer;
 end;

const
  SpKeys: array[0..41] of TSpKeyTable = (
    (Name: 'BS';    Id: VK_BACK),
    (Name: 'Tab';   Id: VK_TAB),
    (Name: 'Enter'; Id: VK_RETURN),
    (Name: 'Esc';   Id: VK_ESCAPE),
    (Name: 'Space'; Id: VK_SPACE),
    (Name: 'PgUp';  Id: VK_PRIOR),
    (Name: 'PgDn';  Id: VK_NEXT),
    (Name: 'End';   Id: VK_END),
    (Name: 'Home';  Id: VK_HOME),
    (Name: 'Left';  Id: VK_LEFT),
    (Name: 'Up';    Id: VK_UP),
    (Name: 'Right'; Id: VK_RIGHT),
    (Name: 'Down';  Id: VK_DOWN),
    (Name: 'Ins';   Id: VK_INSERT),
    (Name: 'Del';   Id: VK_DELETE),
    (Name: 'Num0';  Id: VK_NUMPAD0),
    (Name: 'Num1';  Id: VK_NUMPAD1),
    (Name: 'Num2';  Id: VK_NUMPAD2),
    (Name: 'Num3';  Id: VK_NUMPAD3),
    (Name: 'Num4';  Id: VK_NUMPAD4),
    (Name: 'Num5';  Id: VK_NUMPAD5),
    (Name: 'Num6';  Id: VK_NUMPAD6),
    (Name: 'Num7';  Id: VK_NUMPAD7),
    (Name: 'Num8';  Id: VK_NUMPAD8),
    (Name: 'Num9';  Id: VK_NUMPAD9),
    (Name: 'Num*';  Id: VK_MULTIPLY),
    (Name: 'Num+';  Id: VK_ADD),
    (Name: 'Num-';  Id: VK_SUBTRACT),
    (Name: 'Num.';  Id: VK_DECIMAL),
    (Name: 'Num/';  Id: VK_DIVIDE),
    (Name: 'F1';    Id: VK_F1),
    (Name: 'F2';    Id: VK_F2),
    (Name: 'F3';    Id: VK_F3),
    (Name: 'F4';    Id: VK_F4),
    (Name: 'F5';    Id: VK_F5),
    (Name: 'F6';    Id: VK_F6),
    (Name: 'F7';    Id: VK_F7),
    (Name: 'F8';    Id: VK_F8),
    (Name: 'F9';    Id: VK_F9),
    (Name: 'F10';   Id: VK_F10),
    (Name: 'F11';   Id: VK_F11),
    (Name: 'F12';   Id: VK_F12)
  );

procedure StrToKey(const KeyName: string; var KeyState: TKeyState; var VKey: UINT);
var
  Key, Shift: string;
  i: Integer;
begin
  KeyState := [];

  i := Pos('-', KeyName);
  if i < 2 then
    Key := KeyName
  else begin
    Key := Copy(KeyName, i+1, Length(KeyName)-i);
    Shift := Copy(KeyName, 1, i-1);
    for i := 1 to Length(Shift) do
    begin
      case Shift[i] of
        'A': Include(KeyState, ksAlt);
        'C': Include(KeyState, ksCtrl);
        'S': Include(KeyState, ksShift);
        'W': Include(KeyState, ksWin);
      end;
    end;
  end;

  for i := 0 to High(SpKeys) do
  begin
    if Key = SpKeys[i].Name then
    begin
      VKey := SpKeys[i].Id;
      Exit;
    end;
  end;

  Vkey := VkKeyScan(Key[1]) and $FF;
end;

constructor TSCmd.Create(const CmdList: array of string; const DefaultKeys: array of string;
  FileName: string);
var
  i: Integer;
  Cmd: PCmd;
  s: string;
begin
  FCmdTableCount := High(CmdList)+1;
  GetMem(FCmdTable, SizeOf(TCmd)*FCmdTableCount);
  Initialize(FCmdTable^, FCmdTableCount);
  for i := 0 to FCmdTableCount-1 do
  begin
    FCmdTable[i].Name := CmdList[i];
    FCmdTable[i].Id := IDM_SCMD_FIRST+i;
  end;

  LoadKeymapFromArray(DefaultKeys);
  LoadKeymapFromFile(FileName);
  AccHandle := CreateAcceleratorTable(FKeyList, FKeyListCount);

  for i := 0 to FKeyListCount - 1 do
  begin
    Cmd := IdToCmd(FKeyList[i].cmd);
    if Cmd <> nil then
    begin
      s := AccelToStr(FKeyList[i]);
      if Cmd.KeyName = '' then
        Cmd.KeyName := s
      else if s <> '' then
        Cmd.KeyName := Cmd.KeyName + ', ' + s;
    end;
  end;
end;

destructor TSCmd.Destroy;
begin
  Finalize(FCmdTable^, FCmdTableCount);
  FreeMem(FCmdTable);
  DestroyAcceleratorTable(AccHandle);
  inherited Destroy;
end;

procedure TSCmd.LoadKeymapFromArray(const Keymap: array of string);
var
  i: Integer;
begin
  FKeyListCount := (High(Keymap) + 1) div 2;
  for i := 0 to FKeyListCount - 1 do
    FKeyList[i] := StrToAccel(Keymap[i*2], Keymap[i*2+1]);
end;

procedure TSCmd.LoadKeymapFromFile(const FileName: string);
var
  i, ArgCount: Integer;
  Name, Arg: string;
  ac: TAccel;
  ow: Boolean;
begin
  with TSConfFile.Create(FileName) do
    try
      while GetLine(Name, ArgCount, Arg) and (FKeyListCount <= High(FKeyList)) do
      begin
        ac := StrToAccel(Name, Arg);
        ow := False;
        for i := 0 to FKeyListCount-1 do
        begin
          if (ac.key = FKeyList[i].key) and (ac.fVirt = FKeyList[i].fVirt) then
          begin
            FKeyList[i] := ac; // overwrite
            ow := True;
            Break;
          end;
        end;
        if not ow then
        begin
          FKeyList[FKeyListCount] := StrToAccel(Name, Arg);
          Inc(FKeyListCount);
        end;
      end;
    finally
      Free;
    end;
end;

function TSCmd.StrToAccel(const KeyName: string; const ActionName: string): TAccel;
var
  KeyState: TKeyState;
  VKey: UINT;
begin
  Result.cmd := Id(ActionName);
  StrToKey(KeyName, KeyState, VKey);
  Result.key := VKey;
  Result.fVirt := FVIRTKEY;
  if ksAlt in KeyState then
    Result.fVirt := Result.fVirt or FALT;
  if ksCtrl in KeyState then
    Result.fVirt := Result.fVirt or FCONTROL;
  if ksShift in KeyState then
    Result.fVirt := Result.fVirt or FSHIFT;
end;

function TSCmd.AccelToStr(const Accel: TAccel): string;
var
  i: Integer;
  //ScanCode: Integer;
  //buf: array[0..31] of Char;
begin
  Result := '';
  if (Accel.fVirt and FALT) <> 0 then
    Result := 'Alt+';
  if (Accel.fVirt and FSHIFT) <> 0 then
    Result := Result + 'Shift+';
  if (Accel.fVirt and FCONTROL) <> 0 then
    Result := Result + 'Ctrl+';

  {
  // zL[gL[ǂʂłȂƗpłȂ
  ScanCode := MapVirtualKey(Accel.key, 0);
  if ScanCode <> 0 then
  begin
    GetKeyNameText(ScanCode shl 16 or 0 shl 24, buf, SizeOf(buf)-1);
    Result := Result + buf;
  end;
  }

  for i := 0 to High(SpKeys) do
  begin
    if Accel.key = SpKeys[i].Id then
    begin
      Result := Result + SpKeys[i].Name;
      Exit;
    end;
  end;

  Result := Result + Chr(MapVirtualKey(Accel.key, 2) and $FF);
  {
  if (Accel.key > $2F) and (Accel.key < $5B) then
    Result := Result + Chr(Accel.key)
  else
    Result := '';
  }
end;

function TSCmd.IdToCmd(Id: Integer): PCmd;
begin
  if (Id < IDM_SCMD_FIRST) or (Id > IDM_SCMD_FIRST + FCmdTableCount - 1) then
    Result := nil
  else
    Result := @FCmdTable[Id-IDM_SCMD_FIRST];
end;

function TSCmd.NameToCmd(const Name: string): PCmd;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to FCmdTableCount - 1 do
  begin
    if Name = FCmdTable[i].Name then
    begin
      Result := @FCmdTable[i];
      Break;
    end;
  end;
end;

function TSCmd.Name(Id: Integer): string;
var
  Cmd: PCmd;
begin
  Cmd := IdToCmd(Id);
  if Cmd <> nil then
    Result := Cmd.Name
  else
    Result := '';
end;

function TSCmd.Id(const Name: string): Integer;
var
  Cmd: PCmd;
begin
  Cmd := NameToCmd(Name);
  if Cmd <> nil then
    Result := Cmd.Id
  else
    Result := 0;
end;

{ TSCmdMenu }

constructor TSCmdMenu.Create(Cmd: TSCmd; const MenuItemTable: array of string);
var
  i: Integer;
begin
  FCmd := Cmd;

  FItemsCount := (High(MenuItemTable)+1) div 2;
  GetMem(FItems, SizeOf(TCmdMenuItem)*FItemsCount);
  Initialize(FItems^, FItemsCount);
  for i := 0 to FItemsCount-1 do
  begin
    FItems[i].Caption := MenuItemTable[i*2];
    FItems[i].CmdName := MenuItemTable[i*2+1];
    FItems[i].Enabled := True;
    FItems[i].Checked := False;
  end;
end;

destructor TSCmdMenu.Destroy;
begin
  Finalize(FItems^, FItemsCount);
  FreeMem(FItems);
end;

function TSCmdMenu.MakeMenuHandle: HMENU;
var
  i: Integer;
  Cmd: PCmd;
  CurMenu, SubMenu: HMENU;
  State: Integer;
begin
  Result := CreatePopupMenu;
  CurMenu := Result;
  for i := 0 to FItemsCount-1 do
  begin
    with FItems[i] do
    begin
      State := 0;
      if not Enabled then
        State := MF_GRAYED;
      if Checked then
        State := State or MF_CHECKED;

      if Caption = '-' then
        AppendMenu(CurMenu, MF_SEPARATOR, 0, nil)
      else if CmdName[1] = '{' then
      begin
        SubMenu := CreatePopupMenu;
        AppendMenu(CurMenu, MF_POPUP or State, SubMenu, PChar(Caption));
        CurMenu := SubMenu;
      end
      else if CmdName = '}' then
        CurMenu := Result
      else begin
        Cmd := FCmd.NameToCmd(CmdName);
        if Cmd <> nil then
          with Cmd^ do
            AppendMenu(CurMenu, MF_STRING or State, Id, PChar(Caption+#9+KeyName));
      end;
    end;
  end;
end;

procedure TSCmdMenu.DisableItems(const Names: array of string);
var
  i, j: Integer;
begin
  for i := 0 to FItemsCount-1 do
    for j := 0 to High(Names) do
      if FItems[i].CmdName = Names[j] then
        FItems[i].Enabled := False;
end;

procedure TSCmdMenu.CheckItem(const Name: string; Checked: Boolean);
var
  i: Integer;
begin
  for i := 0 to FItemsCount-1 do
    if FItems[i].CmdName = Name then
      FItems[i].Checked := Checked;
end;

end.
