(* System, WindowsjbgȊOɈˑȂ
 * ŗǂƂSetLengthł͂ȂSetStringg(̂ق悢)
 * ύXtestsutilŃeXg邱
 *)

unit SUtil;

interface

uses Windows;

{ Type Convert Function }

function IStr(Value: Integer): string;
function FStr(const Value: Extended; Digits: Integer): string;
function IZStr(Value, Width: Integer): string;
function SStr(const Args: array of const): string;
function SInt(const S: string): Integer;

{ Strings Function }

function Lower(const S: string): string;
function Upper(const S: string): string;
function LStr(const S: string; Count: Integer): string;
function RStr(const S: string; Count: Integer): string;
function RevStr(const S: string): string;
function StrColl(const S1, S2: string): Integer; // P[lr
function ConvLineBreak(const S: string; const LineBreak: string): string;

{ Path Function }

function IsFile(const Path: string): Boolean;
function IsDir(const Path: string): Boolean;
function AbsPath(const Path: string): string;
procedure SplitDrive(const Path: string; var Drive, Tail: string);
procedure SplitPath(const Path: string; var Head, Tail: string);
function BaseName(const Path: string): string;
function DirName(const Path: string): string;
procedure SplitExt(const Path: string; var Root, Ext: string);
function ExpandVars(const Path: string): string;

{ Operating System Services }

function GetEnv(const Name: string; const Default: string): string;
function SysError(Code: Integer): string;

{ Misc. }

function NumFormat(Value: Integer): string;
function UnitFormat(Value: Integer): string;
function GetLongFileName(const FileName: string): string;
function FileTimeToLocalTimeStr(const FileTime: TFileTime): string;
function GetFindData(const Path: string; var FindData: TWin32FindData): Boolean;
function time_tToFileTime(t: Longint): TFileTime;

function StrToColor(S: string): COLORREF;
function StrToBool(const S: string): Boolean;
function ToLogFont(ArgCount: Integer; Args: array of string): TLogFont;

{ Classes }

type
  TListDir = class(TObject)
  private
    FHandle: THandle;
    FResult: Boolean;
    FFindData: TWin32FindData;
    FIsDir: Boolean;
  public
    constructor Create(const Path: string; Mask: string);
    destructor Destroy; override;
    function Each(var FileName: string): Boolean;
    property IsDir: Boolean read FIsDir;
  end;

  TSFile = class(TObject)
  private
    FFile: File;
    FOpend: Boolean;
    FResultCode: Integer;
    function GetSize: Integer;
  public
    constructor Create(const FileName: string; const Mode: string);
    destructor Destroy; override;
    function Read(var Buffer; Count: Integer): Longint;
    function Write(var Buffer; Count: Integer): Longint;
    property ResultCode: Integer read FResultCode;
    property Size: Integer read GetSize;
  end;

const
  MaxGetEnvLen = 1024; // Ō̃k܂
  MaxExpandVarsLen = 1024; // Ō̃k܂

var
  LeadBytes: set of Char = [];


implementation

procedure InitLeadBytes;
var
  CPInfo: TCPInfo;
  i, j: Integer;
begin
  GetCPInfo(CP_ACP, CPInfo);
  i := 0;
  with CPInfo do
    while ((LeadByte[i] or LeadByte[i+1]) <> 0) and (i<MAX_LEADBYTES) do
    begin
      for j := LeadByte[i] to LeadByte[i+1] do
        Include(LeadBytes, Char(j));
      Inc(i, 2);
    end;
end;


{ Type Convert Function }

function IStr(Value: Integer): string;
begin
  Str(Value, Result);
end;

function FStr(const Value: Extended; Digits: Integer): string;
begin
  Str(Value:0:Digits, Result);
end;

function IZStr(Value, Width: Integer): string;
var
  sign: Boolean;
  p: PChar;
begin
  sign := Value < 0;
  if sign then
  begin
    Value := -Value;
    Dec(Width);
  end;
  Str(Value :Width, Result);
  p := PChar(Result);
  while p^ = ' ' do
  begin
    p^ := '0';
    Inc(p);
  end;
  if sign then
    Result := '-' + Result;
end;

function SStr(const Args: array of const): string;
const
  BoolChars: array[Boolean] of string = ('False', 'True');
var
  i: Integer;
  s: string;
begin
  Result := '';
  for i := 0 to High(Args) do
    with Args[i] do
      case VType of
      	vtInteger:
          begin
            Str(VInteger, s);
            Result := Result + s;
          end;
      	vtBoolean: Result := Result + BoolChars[VBoolean];
        VtChar: Result := Result + VChar;
        //vtExtended: Result := Result + SFStr(VExtended^);
        vtString: Result := Result + VString^;
        vtPChar: Result := Result + VPChar;
        vtObject: Result := Result + VObject.ClassName;
        vtClass: Result := Result + VClass.ClassName;
        vtAnsiString: Result := Result + string(VAnsiString);
        //vtCurrency: Result := Result + CurrToStr(VCurrency^);
        //vtVariant: Result := Result + string(VVariant^);
      end;
end;

function SInt(const S: string): Integer;
var
  code: Integer;
begin
  Val(S, Result, code);
end;


{ String Function }

function Lower(const S: string): string;
begin
  SetString(Result, PChar(S), Length(S));
  CharLowerBuff(PChar(Result), Length(Result));
end;

function Upper(const S: string): string;
begin
  SetString(Result, PChar(S), Length(S));
  CharUpperBuff(PChar(Result), Length(Result));
end;

function LStr(const S: string; Count: Integer): string;
var
  i: Integer;
begin
  if Count >= Length(S) then
  begin
    Result := S;
    Exit;
  end;
  i := 0;
  while i < Count do
  begin
    Inc(i);
    if S[i] in LeadBytes then
      Inc(i);
  end;
  if i > Count then
    Dec(i, 2);
  Result := Copy(S, 1, i);
end;

function RStr(const S: string; Count: Integer): string;
var
  i: Integer;
begin
  if Count >= Length(S) then
  begin
    Result := S;
    Exit;
  end;
  Count := Length(S) - Count;
  i := 0;
  while i < Count do
  begin
    Inc(i);
    if S[i] in LeadBytes then
      Inc(i);
  end;
  Result := Copy(S, i+1, Length(S)-i);
end;

function RevStr(const S: string): string;
var
  i, len: Integer;
begin
  len := Length(S);
  SetString(Result, nil, len);
  i := 0;
  while i < len do
  begin
    Inc(i);
    if S[i] in LeadBytes then
    begin
      Result[len-i] := S[i];
      Result[len-i+1] := S[i+1];
      Inc(i);
    end else
      Result[len-i+1] := S[i];
  end;
end;

function StrColl(const S1, S2: string): Integer;
begin
  Result := lstrcmp(PChar(S1), PChar(S2));
end;

function ConvLineBreak(const S: string; const LineBreak: string): string;
var
  sp, ep, p: PChar;
  lblen, n: Integer;
begin
  lblen := Length(LineBreak);
  SetString(Result, nil, Length(S)+lblen-1);
  sp := @S[1];
  ep := @S[Length(S)+1];
  p := sp;
  n := 1;
  while p < ep do
  begin
    if p^ in [#10, #13] then
    begin
      Move(sp^, Result[n], p-sp);
      Inc(n, p-sp);
      Move(LineBreak[1], Result[n], lblen);
      Inc(n, lblen);
      if (p^ = #13) and ((p+1)^ = #10) then
        Inc(p);
      Inc(p);
      SetLength(Result, Length(Result)+lblen-1);
      sp := p;
    end else
      Inc(p);
  end;
  Move(sp^, Result[n], p-sp);
  Inc(n, p-sp);
  SetLength(Result, n-1);
end;

{ Path Function }

const
  PathSep = '\';
  ExtSep = '.';

function IsFile(const Path: string): Boolean;
var
  r: DWORD;
begin
  r := GetFileAttributes(PChar(Path));
  Result := (r <> $FFFFFFFF) and ((r and FILE_ATTRIBUTE_DIRECTORY) = 0);
end;

function IsDir(const Path: string): Boolean;
var
  r: DWORD;
begin
  r := GetFileAttributes(PChar(Path));
  Result := (r <> $FFFFFFFF) and ((r and FILE_ATTRIBUTE_DIRECTORY) <> 0);
end;

function AbsPath(const Path: string): string;
var
  buf: array [0..MAX_PATH] of Char;
  p: PChar;
  len: Integer;
begin
  len := GetFullPathName(PChar(Path), SizeOf(buf), buf, p);
  if (RStr(buf, 1) = PathSep) and (len > 3) then
    Dec(len);
  SetString(Result, buf, len);
end;

procedure SplitDrive(const Path: string; var Drive, Tail: string);
begin
  if (Length(Path) > 1) and (Path[2] = ':') then
  begin
    Drive := Copy(Path, 1, 2);
    Tail := Copy(Path, 3, Length(Path)-2);
  end else
  begin
    Drive := '';
    Tail := Path;
  end;
end;

procedure SplitPath(const Path: string; var Head, Tail: string);
var
  i, dpos: Integer;
  p: string;
begin
  SplitDrive(Path, Head, p);
  dpos := 0;
  i := 0;
  while i < Length(p) do
  begin
    Inc(i);
    if p[i] in LeadBytes then
      Inc(i)
    else if p[i] = PathSep then
      dpos := i;
  end;
  Head := Head + Copy(p, 1, dpos-1);
  if (dpos = 1) or (dpos = 2) then  // C:\hoge or \\hoge
    Head := Head + PathSep;
  Tail := Copy(p, dpos+1, Length(p)-dpos);
end;

function BaseName(const Path: string): string;
var
  s: string;
begin
  SplitPath(Path, s, Result);
end;

function DirName(const Path: string): string;
var
  s: string;
begin
  SplitPath(Path, Result, s);
end;

procedure SplitExt(const Path: string; var Root, Ext: string);
var
  i: Integer;
  Head, Tail: string;
begin
  SplitPath(Path, Head, Tail);
  for i := Length(Tail) downto 1 do
    if Tail[i] = ExtSep then
    begin
      if (Head <> '') and (RStr(Head, 1) <> PathSep) then
        Head := Head + PathSep;
      Root := Head + Copy(Tail, 1, i-1);
      Ext := Copy(Tail, i, Length(Tail)-i+1);
      Exit;
    end;
  Root := Path;
  Ext := '';
end;

function ExpandVars(const Path: string): string;
var
  buf: array[0..MaxExpandVarsLen-1] of Char;
begin
  // ExpandEnvironmentStringsWin2kƕA9xƃoCgԂ̂
  // Ԃl𗘗płȂ:-(
  ExpandEnvironmentStrings(PChar(Path), buf, SizeOf(buf));
  Result := buf;
end;


{ Operation System Services }

function GetEnv(const Name: string; const Default: string): string;
var
  buf: array[0..MaxGetEnvLen-1] of Char;
begin
  SetString(Result, Buf, GetEnvironmentVariable(PChar(Name), Buf, SizeOf(Buf)));
  if Result = '' then
    Result := Default;
end;

function SysError(Code: Integer): string;
var
  Buf: PChar;
begin
  SetString(Result, Buf, FormatMessage(
    FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_ALLOCATE_BUFFER or $FF,
    nil, Code, (SUBLANG_DEFAULT shl 10) or LANG_NEUTRAL, PChar(@Buf), 0, nil));
  LocalFree(HLOCAL(Buf));
  if Result = '' then
    Result := 'Unknown error: ' + IStr(Code);
end;


{ Misc. }

function GetLongFileName(const FileName: string): string;
var
  h: THandle;
  FindData: TWin32FindData;
  Path: string;
begin
  Result := '';
  if FileName = '' then
    Exit;
  Path := AbsPath(FileName);
  while True do
  begin
    h := FindFirstFile(PChar(Path), FindData);
    if h = INVALID_HANDLE_VALUE then
    begin
      Result := Path + Result;
      Exit;
    end;
    Windows.FindClose(h);

    // '.' 狤L(@Win95a)??
    if string(FindData.cFileName) = '.' then
    begin
      Result := Path + Result;
      Exit;
    end;

    Path := DirName(Path);
    if (Path = '\\') or ((Length(Path) = 3) and (Path[2] = ':')) then
    begin
      Result := Path + FindData.cFileName + Result;
      Exit;
    end;

    Result := '\' + string(FindData.cFileName) + Result;
  end;
end;

function NumFormat(Value: Integer): string;
var
  s: string;
  nf: TNumberFmt;
  buf: array[0..255] of Char;
begin
  nf.NumDigits := 0;
  nf.LeadingZero := 0;
  nf.Grouping := 3;
  nf.lpDecimalSep := '.';
  nf.lpThousandSep := ',';
  nf.NegativeOrder := 1;
  Str(Value, s);
  GetNumberFormat(LOCALE_USER_DEFAULT, 0, PChar(s), @nf, buf, SizeOf(buf));
  Result := buf;
end;

// Explorer̃oCg\Lɐ4(_܂)+Pʕɂ
function UnitFormat(Value: Integer): string;
const
  KiroSize = 1024;
  MegaSize = 1024 * 1024;
  GigaSize = 1024 * 1024 * 1024;
var
  u: Char;
  s: string;
begin
  if Value < KiroSize then
    Str(Value, Result)
  else begin
    if Value >= MegaSize * 1000 then
    begin
      Value := Value div MegaSize;
      u := 'G';
    end
    else if Value >= KiroSize * 1000 then
    begin
      Value := Value div KiroSize;
      u := 'M';
    end else
      u := 'K';
    Value := Value * 100 div KiroSize;
    Str(Value, s);
    if Value >= 10000 then
      Result := s[1]+s[2]+s[3]
    else if Value >= 1000 then
      Result := s[1]+s[2]+'.'+s[3]
    else if Value >= 100 then
      Result := s[1]+'.'+s[2]+s[3]
    else
      Result := '0.'+s[1]+s[2];
    Result := Result + u;
  end;
end;


// FileTime: UTC
function FileTimeToLocalTimeStr(const FileTime: TFileTime): string;
var
  buf: array[0..1023] of Char;
  st: TSystemTime;
  ft: TFileTime;
begin
  if (FileTime.dwLowDateTime = 0) and (FileTime.dwHighDateTime = 0) then
  begin
    Result := '';
    Exit;
  end;
  FileTimeToLocalFileTime(FileTime, ft);
  FileTimeToSystemTime(ft, st);
  GetDateFormat(LOCALE_USER_DEFAULT, DATE_LONGDATE, @st, nil, buf, SizeOf(buf));
  Result := buf;
  GetTimeFormat(LOCALE_USER_DEFAULT, 0, @st, nil, buf, SizeOf(buf));
  Result := Result + ' ' + buf;
end;

function GetFindData(const Path: string; var FindData: TWin32FindData): Boolean;
var
  h: THandle;
begin
  FillChar(FindData, SizeOf(FindData), 0);
  h := FindFirstFile(PChar(Path), FindData);
  Result := h <> INVALID_HANDLE_VALUE;
  if Result then
    Windows.FindClose(h);
end;

function time_tToFileTime(t: Longint): TFileTime;
var
  st: TSystemTime;
  ft: TFileTime;
begin
  st.wYear := 1970;
  st.wMonth := 1;
  st.wDay := 1;
  st.wHour := 0;
  st.wMinute := 0;
  st.wSecond := 0;
  st.wMilliseconds := 0;
  SystemTimeToFileTime(st, ft);
  Comp(Result) := Comp(ft) + t * 10000000.0;
end;

// 1, on, true Ȃ True Ԃ(ignore case)
function StrToBool(const S: string): Boolean;
begin
  Result := (lstrcmpi(PChar(S), '1') = 0) or (lstrcmpi(PChar(S), 'on') = 0)
    or (lstrcmpi(PChar(S), 'true') = 0);
end;

// '#FFFFFF' `̕񂩂 COLORREF l
function StrToColor(S: string): COLORREF;
var
  col, code: Integer;
begin
  if (S <> '') and (S[1] = '#') then
  begin
    S[1] := '$';
    Val(S, col, code);
    Result := ((col shl 16) and $FF0000) or (col and $00FF00) or (col shr 16);
  end else
    Result := COLORREF(-1);
end;

// Height, FontName, BoldFlg, ItalicFlg, CharSet
// ArgsLOGFONTԂ
function ToLogFont(ArgCount: Integer; Args: array of string): TLogFont;
  function StrToInt(const s: string): Integer;
  var
    code: Integer;
  begin
    Val(s, Result, code);
  end;
var
  i: Integer;
begin
  FillChar(Result, SizeOf(Result), 0);
  with Result do
  begin
    lfOutPrecision   := OUT_DEFAULT_PRECIS;
    lfClipPrecision  := CLIP_DEFAULT_PRECIS;
    lfQuality        := DRAFT_QUALITY;
    lfPitchAndFamily := DEFAULT_PITCH;
    lfWeight := FW_NORMAL;
    lfCharSet := DEFAULT_CHARSET;
    for i := 0 to ArgCount - 1 do
    begin
      case i of
        0: lfHeight := StrToInt(Args[i]);
        1: lstrcpyn(lfFaceName, PChar(Args[i]), SizeOf(lfFaceName));
        2: if StrToInt(Args[i]) <> 0 then lfWeight := FW_BOLD;
        3: lfItalic := StrToInt(Args[i]);
        4: lfCharSet := StrToInt(Args[i]);
      end;
    end;
  end;
end;

{ TListDir }

constructor TListDir.Create(const Path: string; Mask: string);
begin
  if Mask = '' then
    Mask := '*.*';
  FHandle := FindFirstFile(PChar(Path+'\'+Mask), FFindData);
  FResult := FHandle <> INVALID_HANDLE_VALUE;
end;

destructor TListDir.Destroy;
begin
  if FHandle <> INVALID_HANDLE_VALUE then
    Windows.FindClose(FHandle);
  inherited;
end;

function TListDir.Each(var FileName: string): Boolean;
begin
  Result := FResult;
  if FResult then
  begin
    FileName := FFindData.cFileName;
    FIsDir := (FFindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) <> 0;
    FResult := FindNextFile(FHandle, FFindData);
  end;
end;


{ TSFile }

constructor TSFile.Create(const FileName: string; const Mode: string);
begin
  FileMode := 0; // read only mode
  AssignFile(FFile, FileName);
  {$I-}
  if Mode = 'w' then
    Rewrite(FFile, 1)
  else
    Reset(FFile, 1);
  {$I+}
  FResultCode := IOResult;
  FOpend := FResultCode = 0;
end;

destructor TSFile.Destroy;
begin
  if FOpend then
    CloseFile(FFile);
  inherited;
end;

function TSFile.GetSize: Integer;
begin
  Result := FileSize(FFile);
end;

function TSFile.Read(var Buffer; Count: Integer): Longint;
begin
  {$I-}
  BlockRead(FFile, Buffer, Count, Result);
  {$I+}
  FResultCode := IOResult;
end;

function TSFile.Write(var Buffer; Count: Integer): Longint;
begin
  {$I-}
  BlockWrite(FFile, Buffer, Count, Result);
  {$I+}
  FResultCode := IOResult;
end;

initialization
  InitLeadBytes;

end.

