{
  DIB to DIB Filter
}

unit in_dib;

interface

uses Windows, npi;

type
  TDibFilter = class(TImageFilter)
  private
    function GetName: string; override;
    function GetPitch(Width, BitCount: Integer): Integer;
    function IsRLE(BmiHeader: PBitmapInfoHeader): Boolean;
    function IsOS2(BmiHeader: PBitmapInfoHeader): Boolean;
    function DecodeRLE(InBuf: PChar; InBufSize: Integer; OutBuf: PChar; OutBufSize: Integer): Integer;
    procedure CoreInfoToBitmapInfo(BmInfo: PBitmapInfo; BmCore: PBitmapCoreInfo);
    procedure GetFormatStr(InBuf: PChar; var ImageMetaInfo: TImageMetaInfo);
    procedure GetBitmapInfo(InBuf: PChar; BitmapInfo: PBitmapInfo; var SrcBpp: Integer);
  public
    function Test(InBuf: PChar; InBufSize: Integer): Boolean; override;
    function GetInfo(InBuf: PChar; InBufSize: Longint; var ImageInfo: TImageInfo;
      ImageMetaInfo: PImageMetaInfo): Integer; override;
    function Decode(InBuf: PChar; InBufSize: Longint; OutBuf: PChar;
      OutBufSize: Longint): Integer; override;
    function Decode2(InBuf: PChar; InBufSize: Longint; OutBuf: PChar; OutBufSize: Longint;
      BitmapInfo: PBitmapInfo): Integer;
    procedure FreeImageMetaInfo(var ImageMetaInfo: TImageMetaInfo); override;
  end;

implementation

{ TDibFilter }

function TDibFilter.GetName: string;
begin
  Result := '(Internal DIB to DIB filter)';
end;

function TDibFilter.GetPitch(Width, BitCount: Integer): Integer;
begin
  Result:= ((Width * BitCount + 31) div 32) * 4;
end;

function TDibFilter.IsRLE(BmiHeader: PBitmapInfoHeader): Boolean;
begin
  Result := (BmiHeader.biCompression = BI_RLE8) or (BmiHeader.biCompression = BI_RLE4);
end;

function TDibFilter.IsOS2(BmiHeader: PBitmapInfoHeader): Boolean;
begin
  Result := BmiHeader.biSize = SizeOf(TBitmapCoreHeader);
end;

function TDibFilter.DecodeRLE(InBuf: PChar; InBufSize: Integer; OutBuf: PChar; OutBufSize: Integer): Integer;
var
  dc: HDC;
  wnd: HWND;
  bmp: HBITMAP;
  BitmapInfo: PBitmapInfo;
  OffBits: DWORD;
  DecBmi: TBitmapInfo;
begin
  wnd := GetDesktopWindow;
  dc := GetDC(wnd);
  OffBits := PBitmapFileHeader(InBuf).bfOffBits;
  BitmapInfo := PBitmapInfo(@InBuf[SizeOf(TBitmapFileHeader)]);
  // BITMAPINFOHEADER  biSizeImage sȏꍇC (todo:CƂʒm)
  if (BitmapInfo.bmiHeader.biSizeImage = 0) and (BitmapInfo.bmiHeader.biCompression <> BI_RGB) then
    BitmapInfo.bmiHeader.biSizeImage := InBufSize - Offbits;
  bmp := CreateDIBitmap(dc, BitmapInfo.bmiHeader, CBM_INIT, PChar(@InBuf[OffBits]),
    BitmapInfo^, DIB_RGB_COLORS);
  try
    DecBmi := BitmapInfo^;
    with DecBmi.bmiHeader do
    begin
      biCompression := BI_RGB;
      biBitCount := 24;
      biSizeImage := 0;
    end;
    if (GetPitch(DecBmi.bmiHeader.biWidth, 24) * DecBmi.bmiHeader.biHeight) > OutBufSize then
    begin
      Result := NPI_MEMNEEDMORE;
      Exit;
    end;
    GetDIBits(dc, bmp, 0, DecBmi.bmiHeader.biHeight, OutBuf, DecBmi, DIB_RGB_COLORS);
    Result := 0;
  finally
    DeleteObject(bmp);
    ReleaseDC(wnd, dc);
  end;
end;

procedure TDibFilter.CoreInfoToBitmapInfo(BmInfo: PBitmapInfo; BmCore: PBitmapCoreInfo);
var
  i: Integer;
begin
  with BmInfo^ do
  begin
    FillChar(bmiHeader, SizeOf(bmiHeader), 0);
    with bmiHeader do
    begin
      biSize := SizeOf(bmiHeader);
      biWidth := BmCore.bmciHeader.bcWidth;
      biHeight := BmCore.bmciHeader.bcHeight;
      biPlanes := BmCore.bmciHeader.bcPlanes;
      biBitCount := BmCore.bmciHeader.bcBitCount;
      biCompression := BI_RGB;
    end;
    if (bmiHeader.biBitCount = 8) or (bmiHeader.biBitCount = 4) then
    begin
      for i := 0 to 1 shl bmiHeader.biBitCount - 1 do
      begin
        bmiColors[i].rgbBlue := BmCore.bmciColors[i].rgbtBlue;
        bmiColors[i].rgbGreen := BmCore.bmciColors[i].rgbtGreen;
        bmiColors[i].rgbRed := BmCore.bmciColors[i].rgbtRed;
      end;
    end;
  end;
end;

procedure TDibFilter.GetFormatStr(InBuf: PChar; var ImageMetaInfo: TImageMetaInfo);
  // 1̃rbg𐔂
  function CountBits(Value: Integer): Integer;
  begin
    Result := 0;
    repeat
      Inc(Result, Value and 1);
      Value := Value shr 1
    until Value = 0;
  end;
var
  s: string;
  BitFields: array[0..2] of DWORD;
begin
  Inc(InBuf, SizeOf(TBitmapFileHeader));
  if IsOs2(PBitmapInfoHeader(InBuf)) then
    s := 'OS/2 Bitmap'
  else begin
    case PBitmapInfoHeader(InBuf).biCompression of
      BI_RGB:  s := 'RGB';
      BI_RLE8: s := 'RLE8';
      BI_RLE4: s := 'RLE4';
      BI_BITFIELDS:
        begin
          Move(PBitmapInfo(InBuf).bmiColors[0], BitFields, SizeOf(BitFields));
          Str(CountBits(BitFields[0])*100+CountBits(BitFields[1])*10+CountBits(BitFields[2]), s);
          s := 'BITFIELDS (' + s + ')';
        end;
    else
      s := 'unknown';
    end;
    if PBitmapInfoHeader(InBuf).biHeight < 0 then
      s := s + ' (topdown)';
    s := 'Windows Bitmap - ' + s;
  end;
  Move(s[1], ImageMetaInfo.FormatName, Length(s)+1); //+1#0
end;

procedure TDibFilter.GetBitmapInfo(InBuf: PChar; BitmapInfo: PBitmapInfo; var SrcBpp: Integer);
var
  OffBits: Integer;
begin
  // bug: BITFIELD555SrcBpp16ƂĂ܂
  OffBits := PBitmapFileHeader(InBuf).bfOffBits;
  Inc(InBuf, SizeOf(TBitmapFileHeader));
  if IsOS2(PBitmapInfoHeader(InBuf)) then
  begin
    CoreInfoToBitmapInfo(BitmapInfo, PBitmapCoreInfo(InBuf));
    SrcBpp := BitmapInfo.bmiHeader.biBitCount;
  end
  else begin
    Move(InBuf^, BitmapInfo^, OffBits - SizeOf(TBitmapFileHeader));
    SrcBpp := BitmapInfo.bmiHeader.biBitCount;
    // RLE24bitRGBɓWJ
    if IsRLE(PBitmapInfoHeader(InBuf)) then
    begin
      with BitmapInfo.bmiHeader do
      begin
        biBitCount := 24;
        biCompression := BI_RGB;
        biSizeImage := 0;
      end;
    end;
  end;
  // BI_RGB  16bit DIB ̏ꍇ RGB555(=15bit) (Microsoft Knowledge Base J028984)
  if (BitmapInfo.bmiHeader.biBitCount = 16) and (BitmapInfo.bmiHeader.biCompression = BI_RGB) then
    SrcBpp := 15
end;

function TDibFilter.Test(InBuf: PChar; InBufSize: Integer): Boolean;
begin
  if InBufSize < (SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfoHeader)) then
  begin
    Result := False;
    Exit;
  end;
  if (InBuf[0] = 'B') and (InBuf[1] = 'M') then
  begin
    Inc(InBuf, SizeOf(TBitmapFileHeader));
    if IsOS2(PBitmapInfoHeader(InBuf)) or
      (PBitmapInfoHeader(InBuf).biSize = SizeOf(TBitmapInfoHeader)) then
    begin
      Result := True;
      Exit;
    end;
  end;
  Result := False;
end;

function TDibFilter.GetInfo(InBuf: PChar; InBufSize: Longint; var ImageInfo: TImageInfo;
  ImageMetaInfo: PImageMetaInfo): Integer;
var
  MaxBmi: Pointer;
begin
  GetMem(MaxBmi, SizeOf(TBitmapInfoHeader)+SizeOf(TRGBQuad)*256);
  GetBitmapInfo(InBuf, MaxBmi, ImageMetaInfo.SrcBitPerPixel);
  with PBitmapInfoHeader(MaxBmi)^ do
  begin
    ImageInfo.Width := biWidth;
    //ImageInfo.Height := Abs(biHeight);
    ImageInfo.Height := biHeight;
    ImageInfo.BitPerPixel := biBitCount;
  end;
  FreeMem(MaxBmi);

  GetFormatStr(InBuf, ImageMetaInfo^);

  Result := 0;
end;

function TDibFilter.Decode(InBuf: PChar; InBufSize: Longint; OutBuf: PChar; OutBufSize: Longint): Integer;
begin
  Result := NPI_NOSUPPORT;
end;

function TDibFilter.Decode2(InBuf: PChar; InBufSize: Longint; OutBuf: PChar; OutBufSize: Longint;
  BitmapInfo: PBitmapInfo): Integer;
var
  SrcBpp: Integer;
begin
  GetBitmapInfo(InBuf, BitmapInfo, SrcBpp);
  if IsRLE(PBitmapInfoHeader(@InBuf[SizeOf(TBitmapFileHeader)])) then
    Result := DecodeRLE(InBuf, InBufSize, OutBuf, OutBufSize)
  else begin
    with BitmapInfo.bmiHeader do
    begin
      if (GetPitch(biWidth, biBitCount) * biHeight) > OutBufSize then
      begin
        Result := NPI_NOSUPPORT;
        Exit;
      end;
    end;
    Move(InBuf[PBitmapFileHeader(InBuf).bfOffBits], OutBuf^, OutBufSize);
    Result := 0;
  end;
end;

procedure TDibFilter.FreeImageMetaInfo(var ImageMetaInfo: TImageMetaInfo);
begin
end;

end.

