{
  Spi adapter
}

unit in_spi;

interface

uses Windows, npi;

const
  SpiErrMsg: array[0..10] of string = (
    '',
    'user cancel',
    'unknown format',
    'broken data',
    'allocate memory error',
    'memory error',
    'file read error',
    'cannot create window',
    'internal error',
    'file write error',
    'end of file error'
  );

type
  TSpiFilter = class(TImageFilter)
  private
    FHandle: THandle;
    function GetName: string; override;
    function GetFileName: string;
  public
    function Test(InBuf: PChar; InBufSize: Integer): Boolean; override;
    function GetInfo(InBuf: PChar; InBufSize: Longint; var ImageInfo: TImageInfo;
      ImageMetaInfo: PImageMetaInfo): Integer; override;
    function Decode(InBuf: PChar; InBufSize: Longint; OutBuf: PChar;
      OutBufSize: Longint): Integer; override;
    function Decode2(InBuf: PChar; InBufSize: Longint; var HOutBuf: HLOCAL;
      BitmapInfo: PBitmapInfo): Integer;
    function IsArchivePlugin: Boolean;
    procedure FreeImageMetaInfo(var ImageMetaInfo: TImageMetaInfo); override;
    function Load(const AFileName: string): Boolean;
    procedure UnLoad;
    function ConfigDlg(hWnd: HWND): Integer;
    property FileName: string read GetFileName;
    property Handle: THandle read FHandle;
  end;

  TSpiArchiver = class(TObject)
  private
    FHandle: THandle;
    FArcFileName: string;
    FResultCode: Integer;
    FInfoList: PFileInfo;
    function GetInfoList: PFileInfo;
  public
    procedure Open(const ArcFileName: string);
    procedure Close;
    procedure Extract(const FileInfo: TFileInfo; OutBuf: PChar; OutBufSize: Longint);
    property InfoList: PFileInfo read GetInfoList;
    property Handle: THandle read FHandle write FHandle;
    property ResultCode: Integer read FResultCode;
  end;

function GetSpiPath: string;
function SafeIsArchivePlugin(const Filter: TImageFilter): Boolean;

implementation

uses spi_api;

// Susie plug-in ̏ꏊ擾
function GetSpiPath: string;
var
  Buf: array[0..MAX_PATH] of Char;
  size: DWORD;
  hkResult: HKEY;
begin
  FillChar(Buf, SizeOf(Buf), 0);
  Result := '';
  if RegOpenKeyEx(HKEY_CURRENT_USER, 'Software\Takechin\Susie\Plug-in', 0,
    KEY_QUERY_VALUE, hkResult) = ERROR_SUCCESS then
  begin
    Size := SizeOf(Buf);
    if RegQueryValueEx(hkResult, 'Path', nil, nil, PByte(@Buf), @Size) = ERROR_SUCCESS then
      SetString(Result, Buf, Size-1); // -1null terminate
    RegCloseKey(hkResult);
  end;
end;

function SafeIsArchivePlugin(const Filter: TImageFilter): Boolean;
begin
  if Filter is TSpiFilter then
    Result := (Filter as TSpiFilter).IsArchivePlugin
  else
    Result := False;
end;

{ TSpiFilter }

function ProgressCallback(nNum, nDemon: Integer; lData : Longint): Integer; stdcall;
begin
  Result := 0;
end;

function TSpiFilter.GetName: string;
var
  GetPluginInfo: TGetPluginInfo;
  Buf: array[0..255] of Char;
begin
  GetPluginInfo := GetProcAddress(FHandle, 'GetPluginInfo');
  if Assigned(GetPluginInfo) then
  begin
    GetPluginInfo(1, Buf, SizeOf(Buf));
    Result := Buf;
  end else
    Result := '';
end;

function TSpiFilter.GetFileName: string;
var
  Buf: array[0..MAX_PATH] of Char;
begin
  GetModuleFileName(FHandle, Buf, SizeOf(Buf));
  Result := Buf;
end;

function TSpiFilter.Test(InBuf: PChar; InBufSize: Integer): Boolean;
var
  IsSupported: TIsSupported;
begin
  // InBufSize͎gȂ
  IsSupported := GetProcAddress(FHandle, 'IsSupported');
  Result := Assigned(IsSupported) and (IsSupported('', DWORD(InBuf)) <> 0);
end;

function TSpiFilter.GetInfo(InBuf: PChar; InBufSize: Longint; var ImageInfo: TImageInfo;
  ImageMetaInfo: PImageMetaInfo): Integer;
var
  GetPictureInfo: TGetPictureInfo;
  PicInfo: TPictureInfo;
  r: Integer;
  p: PChar;
  s: string;
begin
  FillChar(ImageMetaInfo^, SizeOf(ImageMetaInfo^), 0);
  GetPictureInfo := GetProcAddress(FHandle, 'GetPictureInfo');
  if not Assigned(GetPictureInfo) then
  begin
    Result := NPI_NOSUPPORT;
    Exit;
  end;

  FillChar(PicInfo, SizeOf(PicInfo), 0); // PicInfoSZbgȂvOC΍
  r := GetPictureInfo(InBuf, InBufSize, 1, PicInfo);
  if r <> 0 then
  begin
    Result := r; // Ƃ肠spĩG[R[ĥ܂ܕԂ
    Exit;
  end;

  with PicInfo do
  begin
    ImageInfo.Width := width;
    ImageInfo.Height := height;
    ImageMetaInfo.SrcBitPerPixel := colorDepth;
    if hInfo <> 0 then
    begin
      p := LocalLock(hInfo);
      s := Copy(p, 1, LocalSize(hInfo)-1);
      GetMem(ImageMetaInfo.Comment, Length(s)+1); // +1#0
      if s <> '' then
        Move(s[1], ImageMetaInfo.Comment^, Length(s)+1);
      LocalUnLock(hInfo);
      LocalFree(hInfo);
    end;
  end;
  Result := 0;
end;

function TSpiFilter.Decode(InBuf: PChar; InBufSize: Longint; OutBuf: PChar; OutBufSize: Longint): Integer;
begin
  Result := NPI_NOSUPPORT;
end;

function TSpiFilter.Decode2(InBuf: PChar; InBufSize: Longint; var HOutBuf: HLOCAL;
  BitmapInfo: PBitmapInfo): Integer;
var
  GetPicture: TGetPicture;
  HBInfo: HLOCAL;
  r, size: Integer;
begin
  GetPicture := GetProcAddress(FHandle, 'GetPicture');
  if not Assigned(GetPicture) then
  begin
    Result := NPI_NOSUPPORT; // todo: G[R[hȂ
    Exit;
  end;
  // ProgressCallbacknilƗvOĈŃ_~[֐n
  r := GetPicture(InBuf, InBufSize, 1, HBInfo, HOutBuf, ProgressCallback, 0);
  if r <> 0 then
  begin
    Result := r; // Ƃ肠spĩG[R[ĥ܂ܕԂ
    Exit;
  end;

  size := LocalSize(HBInfo);
  if size > (SizeOf(TBitmapInfoHeader)+SizeOf(TRGBQuad)*259) then
    size := SizeOf(TBitmapInfoHeader)+SizeOf(TRGBQuad)*256;
  Move((LocalLock(HBInfo))^, BitmapInfo^, size);
  LocalUnLock(HBInfo);
  LocalFree(HBInfo);
  Result := 0;
end;

function TSpiFilter.IsArchivePlugin: Boolean;
var
  PluginVersion: array[0..15] of Char;
  GetPluginInfo: TGetPluginInfo;
begin
  FillChar(PluginVersion, SizeOf(PluginVersion), 0);
  GetPluginInfo := GetProcAddress(FHandle, 'GetPluginInfo');
  if Assigned(GetPluginInfo) then
    GetPluginInfo(0, PluginVersion, SizeOf(PluginVersion));
  Result := PluginVersion = '00AM';
end;

procedure TSpiFilter.FreeImageMetaInfo(var ImageMetaInfo: TImageMetaInfo);
begin
  if ImageMetaInfo.Comment = nil then
    Exit;
  FreeMem(ImageMetaInfo.Comment);
end;

function TSpiFilter.Load(const AFileName: string): Boolean;
begin
  FHandle := LoadLibrary(PChar(AFileName));
  Result := FHandle <> 0;
end;

procedure TSpiFilter.UnLoad;
begin
  if FHandle <> 0 then
  begin
    FreeLibrary(FHandle);
    FHandle := 0;
  end;
end;

function TSpiFilter.ConfigDlg(hWnd: HWND): Integer;
var
  ConfDlg: TConfigurationDlg;
begin
  ConfDlg := GetProcAddress(FHandle, 'ConfigurationDlg');
  if Assigned(ConfDlg) then
    Result := ConfDlg(hWnd, 1)
  else
    Result := -1;
end;

{ TSpiArchiver }

procedure TSpiArchiver.Open(const ArcFileName: string);
begin
  FArcFileName := ArcFileName;
end;

procedure TSpiArchiver.Close;
begin
  FArcFileName := '';
  if FInfoList <> nil then
  begin
    FreeMem(FInfoList);
    FInfoList := nil;
  end;
end;

function TSpiArchiver.GetInfoList: npi.PFileInfo;
var
  GetArchiveInfo: TGetArchiveInfo;
  HInfo: HLOCAL;
  SpiFileInfo: spi_api.PFileInfo;
  SpiFileInfoCount: Integer;
  InfoListSize: Integer;
  FileInfo: npi.PFileInfo;
  s: string;
begin
  Result := nil;
  if FArcFileName = '' then
    Exit;
  if FInfoList <> nil then
    FreeMem(FInfoList);

  GetArchiveInfo := GetProcAddress(FHandle, 'GetArchiveInfo');
  FResultCode := GetArchiveInfo(PChar(FArcFileName), 0, 0, HInfo);
  // lhasad.spi 0.7͐ɂ炸unknown format(=2)ԂĂ
  // IsSupportedɐĂȏAunknown format̉\͒Ⴂ̂
  // I0ɏ
  if FResultCode = 2 then
    FResultCode := 0;
  if FResultCode <> 0 then
    Exit;

  SpiFileInfo := LocalLock(HInfo);
  if SpiFileInfo = nil then
    Exit;
  SpiFileInfoCount := LocalSize(HInfo) div SizeOf(spi_api.TFileInfo) + 1;

  InfoListSize := SizeOf(npi.TFileInfo) * SpiFileInfoCount;
  GetMem(FInfoList, InfoListSize);
  FillChar(FInfoList^, InfoListSize, 0);
  FileInfo := FInfoList;

  while SpiFileInfo.method[0] <> #0 do
  begin
    if SpiFileInfo.filesize = 0 then
      FileInfo.FileType := FTYPE_DIR;
    s := SpiFileInfo.Path + SpiFileInfo.filename;
    if Length(s) > (SizeOf(FileInfo.Name)-1) then
      Continue;
    Move(s[1], FileInfo.Name, Length(s)+1); // #0܂ŃRs[
    FileInfo.MTime := SpiFileInfo.timestamp;
    FileInfo.Offset := SpiFileInfo.position;
    FileInfo.FileSize := SpiFileInfo.filesize;
    FileInfo.CompSize := SpiFileInfo.compsize;
    Inc(SpiFileInfo);
    Inc(FileInfo);
  end;

  LocalUnlock(HInfo);
  LocalFree(HInfo);
  Result := FInfoList;
end;

procedure TSpiArchiver.Extract(const FileInfo: npi.TFileInfo; OutBuf: PChar; OutBufSize: Longint);
var
  GetFile: TGetFile;
  HDst: HLOCAL;
  p: PChar;
  size: Integer;
begin
  GetFile := GetProcAddress(FHandle, 'GetFile');
  FResultCode := GetFile(PChar(FArcFileName), FileInfo.Offset, PChar(@HDst), $100, nil, 0);
  if FResultCode <> 0 then
    Exit;
  p := LocalLock(HDst);
  size := LocalSize(HDst);
  if size > OutBufSize then
    size := OutBufSize;
  Move(p^, OutBuf^, size);
  LocalUnLock(HDst);
  LocalFree(HDst);
end;

end.
