(*
    NkV - Graphic Viewer
    Copyright (C) 1998-2004 Nobuyuki Honda (Ferdia) <ferdia@imasy.or.jp>

    Compiler: Delphi 3.01a
*)

program NkV;

uses
  Windows, Messages, ShellAPI, CommDlg,
  SUtil, SForm, SList, SConfFile, SCmd,
  NkdtProf, NkVImage, ResizeImage, in_spi, npi, Command;

{$R *.RES}
{$R nkvres.res}
{$I nkv.inc}

const
  AppName = 'NkV';
  Version = '0.4 (2004-07-20)';
  Copyright = 'Copyright (C) 1999-2004 Nobuyuki Honda (Ferdia)';
  License = 'This software is provided under the zlib/libpng license.';

type
  PFileItem = ^TFileItem;
  TFileItem = record
    Name: string;
    IsImage: Boolean;
    FileInfo: TFileInfo;
  end;

  TFileList = class(TObject)
  private
    FItems: TSList;
    FArcFileName: string;
    FArchiver: TSpiArchiver;
    function Get(Index: Integer): PFileItem;
    function GetItemCount: Integer;
  public
    CurPos: Integer;
    constructor Create;
    destructor Destroy; override;
    procedure AddItem(const Item: TFileItem);
    procedure Sort;
    procedure Clear;
    procedure LoadFromDirectory(const Path: string; const CurFileName: string);
    procedure LoadFromArchive(const FileName: string);
    property Items[Index: Integer]: PFileItem read Get; default;
    property ItemCount: Integer read GetItemCount;
    property ArcFileName: string read FArcFileName write FArcFileName;
    property Archiver: TSpiArchiver read FArchiver;
  end;

  TNkVMainForm = class(TSForm)
  public
    Left, Top: Integer;
    NormalCursor: HCURSOR;
    GrabCursor: HCURSOR;
    Stretch: Boolean;
    ShowColorValue: Boolean;
    ScaleFilter: Boolean;
    ShowGrid: Boolean;
    DispWidth: Integer;         // \
    DispHeight: Integer;        // \
    Zoom: Integer;              // g嗦(%)
    FileName: string;
    FileCTime: TFileTime;       // t@C쐬
    FileATime: TFileTime;       // t@CŏIANZX
    FileMTime: TFileTime;       // t@CŏIݓ
    ArcFileName: string;
    oldpt: TPoint; /////
    procedure InitParams; override;
    procedure ShowAbout;
    procedure ShowPluginList;
    procedure ShowPluginConfig(const PluginFileName: string);
    procedure SetWallPaper(TileFlg: Boolean);
    procedure SetTitle;
    procedure FitWindow;
    procedure ZoomImage(Value: Integer);
    procedure ZoomInOut(OutFlg: Boolean);
    procedure PopupMenu(const pt: TPoint);
    procedure ScrollImage(x, y: Integer);
    function LoadFile(const FileName: string): Boolean;
    function ReadNextFileFromArchive(const FileList: TFileList; PrevFlag: Boolean): Integer;
    function ReadNewFile(const FileName: string; LastFileInArc: Boolean): Integer;
    function ReadFileFromArchive(const FileList: TFileList; Pos: Integer): Integer;
    procedure ReadNextFile(ForceNextFile: Boolean; PrevFlag: Boolean);
    procedure UpdateViewState(OldImageWidth, OldImageHeight: Integer);
    procedure LoadErrMsgBox(Code: Integer; FileName: string);
    procedure ErrMsgBox(const S: string; const Args: array of string);
    procedure CopyFileNameToClipBoard;
    procedure DoCmd(const CmdName: string);
    procedure SaveAs;
    procedure ClientToImage(var pt: TPoint);
    procedure DrawGrid(dc: HDC; l, t, w, h: Integer; Bits: Pointer);

    procedure WMMouseWheel(var Message: TMessage); message WM_MOUSEWHEEL;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
    procedure WMContextMenu(var Message: TMessage); message WM_CONTEXTMENU;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SETCURSOR;
    procedure WMDropFiles(var Message: TWMDropFiles); message WM_DROPFILES;
    procedure WMLButtonDown(var Message: TWMLButtonDown); message WM_LBUTTONDOWN;
    procedure WMLButtonUp(var Message: TWMLButtonUp); message WM_LBUTTONUP;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message WM_LBUTTONDBLCLK;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
  end;

  TPicInfoDlg = class(TSModelessDlg)
    procedure SetItems;
    procedure WMInitDialog(var Message: TWMInitDialog); message WM_INITDIALOG;
    procedure WMCommand(var Message: TWMCommand); message WM_COMMAND;
    procedure WMDrawItem(var Message: TWMDrawItem); message WM_DRAWITEM;
  end;

const
  IDM_PLUGINSETTING_FIRST = 4000;
  IDM_PLUGINSETTING_LAST = IDM_PLUGINSETTING_FIRST + 999;

  Scroll = 16; // J[\gp̃XN[

  ZoomTable: array [0..15] of Integer =
    (12, 25, 50, 66, 100, 125, 150, 200, 300, 400, 500, 600, 700, 800, 1200, 1600);

  WallpaperFile = 'wallpaper.bmp';
  PluginListFile = 'pluginlist';
  KeyMapFile = 'keymap';
  ConfigFile = 'config';

  // -1ȏ̒l͗\
  ErrNotFoundInArchive = -2; // = nilfArchive
  ErrNotFoundInDirectory = -4;

  MsgNotFoundInArchive = '%1%n%nLoadable file not found in archive file.';
  MsgNotFoundInDirectory = '%1%n%nLoadable file not found in directory.';
  MsgDecodeError = '%1%n%n%2 (code=%3)%nPlug-in: %4';
  MsgUnknownFormatError = '%1%n%nUnknown format.';
  MsgPluginLoadError = '%1%n%nLoad failed.';
  MsgPluginConfDlgNotImplement = '%1%n%nConfiguration Dialog is not implemented.';
  MsgSetWallpaperError = 'Cannot set as wallpaper.%n%n%1%n%n';
  MsgSaveAsError = 'Cannot save as file.%n%n%1%n%n';

var
  ProfDir: string;

  MainForm: TNkVMainForm;
  PicInfoDlg: TPicInfoDlg;
  Image: TNkVImage;
  Cmd: TSCmd;
  FileList: TFileList;
  ArcFileList: TFileList;

  InfoDlg: HWND;
  InfoDlgIcon: HICON;

  // Option
  KeepViewPositon: Boolean = False;

function MakePluginListFile: Integer;
var
  s, path: string;
  FileName: string;
begin
  path := GetSpiPath;
  if path = '' then
    path := DirName(ParamStr(0));

  s := '';
  with TListDir.Create(path, '*.spi') do
    try
      while Each(FileName) do
        s := s + path + '\' + FileName + #13#10;
    finally
      Free;
    end;

  if s = '' then
  begin
    Result := 0;
    Exit;
  end;

  with TSFile.Create(ProfDir + PluginListFile, 'w') do
    try
      if ResultCode = 0 then
        Write(s[1], Length(s));
      Result := ResultCode;
    finally
      Free;
    end;
end;

function LoadPluginListFile(const FileName: string): Integer;
var
  Name, Arg: string;
  ArgCount: Integer;
begin
  Result := 0;

  if not IsFile(FileName) then
    Result := MakePluginListFile;

  with TSConfFile.Create(FileName) do
    try
      while GetLine(Name, ArgCount, Arg) do
        Image.FilterList.Add(Name, Arg);
    finally
      Free;
    end;
end;

var
  FormList: TSList;
function EnumWindowsProc(hWnd: HWND; lParam: LPARAM): BOOL; stdcall;
var
  buf: array[0..255] of Char;
  AClassName: string;
begin
  GetClassName(hWnd, buf, SizeOf(buf));
  AClassName := MainForm.ClassName;
  if buf = AClassName then
    FormList.Add(Pointer(hWnd));
  Result := True;
end;

function Compare(Item1, Item2: Pointer): Integer;
begin
  if Integer(Item1) > Integer(Item2) then
    Result := 1
  else if Integer(Item1) < Integer(Item2) then
    Result := -1
  else
    Result := 0;
end;

procedure ActivateNextForm(Prev: Boolean);
var
  i, cur: Integer;
begin
  FormList := TSList.Create;
  EnumWindows(@EnumWindowsProc, 0);
  FormList.Sort(Compare);
  cur := 0;
  for i := 0 to FormList.Count-1 do
  begin
    if HWND(FormList[i]) = MainForm.Handle then
    begin
      cur := i;
      Break;
    end;
  end;
  if Prev then
  begin
    Dec(cur);
    if cur < 0 then
      cur := FormList.Count - 1;
  end else
  begin
    Inc(cur);
    if cur = FormList.Count then
      cur := 0;
  end;
  SetForegroundWindow(HWND(FormList[cur]));
  FormList.Free;
end;

{ TFileList }

constructor TFileList.Create;
begin
  FArchiver := TSpiArchiver.Create;
  FItems := TSList.Create;
end;

destructor TFileList.Destroy;
begin
  Clear;
  FItems.Free;
  FArchiver.Close;
  FArchiver.Free;
end;

procedure TFileList.Clear;
var
  i: Integer;
begin
  for i := 0 to FItems.Count - 1 do
    Dispose(PFileItem(FItems[i]));
  FItems.Clear;
end;

function TFileList.Get(Index: Integer): PFileItem;
begin
  Result := PFileItem(FItems[Index]);
end;

function TFileList.GetItemCount: Integer;
begin
  Result := FItems.Count;
end;

procedure TFileList.AddItem(const Item: TFileItem);
var
  p: PFileItem;
begin
  New(p);
  p^ := Item;
  FItems.Add(p);
end;

function FileListCompare(Item1, Item2: Pointer): Integer;
begin
  Result := StrColl(PFileItem(Item1).Name, PFileItem(Item2).Name);
end;

procedure TFileList.Sort;
begin
  FItems.Sort(FileListCompare);
end;

procedure TFileList.LoadFromDirectory(const Path: string; const CurFileName: string);
var
  Item: TFileItem;
  i: Integer;
begin
  Clear;

  with  TListDir.Create(Path, '') do
    try
      while Each(Item.Name) do
      begin
        if IsDir then
          Continue;
        Item.IsImage := True; // t@C͉摜t@CƂ݂Ȃ
        AddItem(Item);
      end;
    finally
      Free;
    end;

  Sort;
  // ̃t@C̃Xg̈ʒuCurPosɃZbg
  for i := 0 to FItems.Count - 1 do
  begin
    if Items[i].Name = CurFileName then
    begin
      CurPos := i;
      Break;
    end;
  end;
end;

procedure TFileList.LoadFromArchive(const FileName: string);
var
  FileInfo: PFileInfo;
  FileItem: TFileItem;
begin
  FArchiver.Open(FileName);
  try
    FileInfo := FArchiver.InfoList;
    if FArchiver.ResultCode <> 0 then
    begin
      //SpiErrorMsgBox(FileName, SPIError);
      Exit;
    end;

    Clear;
    while FileInfo.Name <> '' do
    begin
     if FileInfo.FileType = FTYPE_FILE then
     begin
       FileItem.Name := FileInfo.Name;
       FileItem.IsImage := True;
       FileItem.FileInfo := FileInfo^;
       AddItem(FileItem);
     end;
     Inc(FileInfo);
    end;
  finally
    FArchiver.Close;
  end;

  Sort;
  FArcFileName := FileName;
end;

{ TPicInfoDlg }

procedure TPicInfoDlg.SetItems;
  function BytesFormat(Value: Integer): string;
  begin
    Result := UnitFormat(Value);
    case Result[Length(Result)] of
      'G': Result := Copy(Result, 1, Length(Result)-1) + ' GB';
      'M': Result := Copy(Result, 1, Length(Result)-1) + ' MB';
      'K': Result := Copy(Result, 1, Length(Result)-1) + ' KB';
    else
      Result := Result + ' bytes';
    end;
  end;

  function GetIcon(const FileName: string): HICON;
  var
    id: Word;
  begin
    id := 1; // id = 1ɂbmpŃTlCACRł͂Ȃʂ̃ACRۂ
    Result := ExtractAssociatedIcon(HInstance, PChar(FileName), id);
  end;
var
  Colors, dir, fname: string;
begin
  with MainForm do
  begin
    if FileName = '' then
    begin
      dir := '';
      fname := '<STDIN>';
      InfoDlgIcon := 0;
    end
    else if ArcFileName = '' then
    begin
      dir := DirName(FileName);
      fname := BaseName(FileName);
      InfoDlgIcon := GetIcon(FileName);
    end else
    begin
      dir := ArcFileName;
      fname := FileName;
      InfoDlgIcon := GetIcon(ArcFileName);
    end;
  end;

  if Image.BitPerPixelOrg < 32 then
    Colors := NumFormat(1 shl Image.BitPerPixelOrg)
  else
    Colors := '';

  Items[IDC_FILENAME].Text := fname;
  Items[IDC_LOCATION].Text := dir;
  Items[IDC_SIZE].Text := SStr([Image.Width, 'x', Image.Height]);
  Items[IDC_COLORS].Text := SStr([Colors, ' (', Image.BitPerPixelOrg, ' bit)']);
  Items[IDC_FORMAT].Text := Image.FormatName;
  Items[IDC_FILESIZE].Text := BytesFormat(Image.FileSize)+' ('+NumFormat(Image.FileSize)+' bytes)';
  Items[IDC_CREATED].Text := FileTimeToLocalTimeStr(MainForm.FileCTime);
  Items[IDC_MODIFIED].Text := FileTimeToLocalTimeStr(MainForm.FileMTime);
  Items[IDC_ACCESSED].Text := FileTimeToLocalTimeStr(MainForm.FileATime);
  Items[IDC_MEMORY].Text := BytesFormat(Image.MemSize)+' ('+NumFormat(Image.MemSize)+' bytes)';
  Items[IDC_PLUGIN].Text := Image.FilterName;
  Items[IDC_PLUGINFILE].Text := Image.FilterFilename;
  Items[IDC_TEXTINFOLABEL].Enabled := Image.Memo <> '';
  Items[IDC_TEXTINFO].Text := Image.Memo;
  InvalidateRect(Items[IDC_ICON].Handle, nil, False);
end;

procedure TPicInfoDlg.WMInitDialog(var Message: TWMInitDialog);
begin
  SetItems;
  Message.Result := 1;
end;

procedure TPicInfoDlg.WMCommand(var Message: TWMCommand);
begin
  with Message do
    case ItemID of
      IDOK, IDCANCEL:
        begin
          Result := 1;
          Close;
        end;
    end;
end;

procedure TPicInfoDlg.WMDrawItem(var Message: TWMDrawItem);
begin
  with Message do
    if Ctl = IDC_ICON then
      with DrawItemStruct^ do
      begin
        FillRect(hDC, rcItem, GetSysColorBrush(COLOR_3DFACE));
        DrawIcon(hDC, 0, 0, InfoDlgIcon);
        Result := 1;
      end;
end;

{ TNkVMainForm }

procedure TNkVMainForm.InitParams;
begin
  inherited;
  Cursor := 0;
  ClassStyle := CS_HREDRAW or CS_VREDRAW or CS_DBLCLKS;
  ExStyle := WS_EX_ACCEPTFILES;
  Brush := 0;

  NormalCursor := LoadCursor(HInstance, 'HANDOPEN');
  GrabCursor := LoadCursor(HInstance, 'HANDHOLD');
end;

procedure TNkVMainForm.ErrMsgBox(const S: string; const Args: array of string);
begin
  FormatMsgBox(Handle, AppName, S, Args, MB_ICONEXCLAMATION);
end;

procedure TNkVMainForm.UpdateViewState(OldImageWidth, OldImageHeight: Integer);
begin
  if (not KeepViewPositon) or (Image.Width <> OldImageWidth) or
    (Image.Height <> OldImageHeight) then
  begin
    Left := 0;
    Top := 0;
  end;
  if Zoom = 0 then
    Zoom := 100;

  DispWidth := Image.Width * Zoom div 100;
  DispHeight := Image.Height * Zoom div 100;
  if not (Stretch or FullScreen or Maximized) then
    FitWindow;
  if IsWindow(InfoDlg) then
    PicInfoDlg.SetItems;
  Invalidate;
  Update;
  SetTitle;
end;

procedure TNkVMainForm.SaveAs;
var
  SaveFileName, Ext: string;
  r: Integer;
begin
  SplitExt(BaseName(FileName), SaveFileName, Ext);
  with TSSaveDlg.Create(Handle, ['All Files', '*.*'], SaveFileName + '.bmp') do
    try
      if ShowModal(SaveFileName) then
      begin
        r := Image.SaveToFile(SaveFileName);
        if r <> 0 then
          ErrMsgBox(MsgSaveAsError+SysError(r), [SaveFileName]);
      end;
    finally
      Free;
    end;
end;

procedure TNkVMainForm.SetWallPaper(TileFlg: Boolean);
var
  WPFileName: string;
  r: Integer;
  hkResult: HKEY;
  data: array[0..MAX_PATH] of Char;
begin
  WPFileName := ProfDir + WallpaperFile;
  r := Image.SaveToFile(WPFileName);
  if r <> 0 then
  begin
    ErrMsgBox(MsgSetWallpaperError+SysError(r), [WPFileName]);
    Exit;
  end;

  if TileFlg then
    data := '1'
  else
    data := '0';
  if RegOpenKeyEx(HKEY_CURRENT_USER, 'Control Panel\desktop', 0, KEY_WRITE, hkResult) = ERROR_SUCCESS then
  begin
    RegSetValueEx(hkResult, 'TileWallPaper', 0, REG_SZ, @data, 1);
    RegCloseKey(hkResult);
  end;

  SystemParametersInfo(SPI_SETDESKWALLPAPER, 0, PChar(WPFileName),
    SPIF_UPDATEINIFILE or SPIF_SENDWININICHANGE);
end;

function TNkVMainForm.ReadFileFromArchive(const FileList: TFileList; Pos: Integer): Integer;
var
  oldCursor: HCURSOR;
  w, h: Integer;
  FileItem: PFileItem;
  pTmp: PChar;
begin
  w := Image.Width;
  h := Image.Height;

  FileItem := FileList[Pos];

  if not FileItem.IsImage then
  begin
    Result := nilfFilterNotFound;
    Exit;
  end;

  oldCursor := SetCursor(LoadCursor(0, IDC_WAIT));
  try
    FileList.Archiver.Open(FileList.ArcFileName);
    GetMem(pTmp, FileItem.FileInfo.FileSize);
    try
      FileList.Archiver.Extract(FileItem.FileInfo, pTmp, FileItem.FileInfo.FileSize);
      FileList.Archiver.Close;
      if FileList.Archiver.ResultCode = 0 then
        Result := Image.LoadFromMemory(pTmp, FileItem.FileInfo.FileSize)
      else begin
        Result := FileList.Archiver.ResultCode;
        Exit;
      end;
    finally
      FreeMem(pTmp);
    end;
  finally
    SetCursor(oldCursor);
  end;

  if Result = 0 then
  begin
    FileName := FileItem.Name;
    ArcFileName := FileList.ArcFileName;
    FillChar(FileCTime, SizeOf(FileCTime) ,0);
    FileMTime := time_tToFileTime(FileItem.FileInfo.MTime);
    FillChar(FileATime, SizeOf(FileATime) ,0);
    UpdateViewState(w, h);
  end else
  begin
    //SPIErrorMsgBox(FileList.ArcFileName+'/'+FileItem.Name, Image.SPIError);
    FileItem.IsImage := False;
  end;
end;


function TNkVMainForm.ReadNextFileFromArchive(const FileList: TFileList; PrevFlag: Boolean): Integer;
var
  st, ed, d, i: Integer;
begin
  Result := -1;
  if PrevFlag then
  begin
    st := FileList.CurPos - 1;
    ed := 0;
    d := -1;
    if ed > st then Exit;
  end else
  begin
    st := FileList.CurPos + 1;
    ed := FileList.ItemCount - 1;
    d := 1;
    if ed < st then Exit;
  end;

  i := st;
  while True do
  begin
    Result := ReadFileFromArchive(FileList, i);
    if Result = 0 then
    begin
      FileList.CurPos := i;
      Break;
    end;
    if i = ed then
      Break;
    Inc(i, d);
  end;
end;

function TNkVMainForm.LoadFile(const FileName: string): Boolean;
var
  TmpFileList: TFileList;
  TmpTmpFileList: TFileList;
  i, r: Integer;
begin
  if IsDir(FileName) then
  begin
    r := ErrNotFoundInDirectory;
    TmpFileList := TFileList.Create;
    try
      TmpFileList.LoadFromDirectory(FileName, '');
      for i := 0 to TmpFileList.ItemCount - 1 do
      begin
        r := ReadNewFile(FileName + '\' + TmpFileList[i].Name, False);
        if r = 0 then
        begin
          TmpTmpFileList := FileList;
          FileList := TmpFileList;
          TmpFileList := TmpTmpFileList;
          FileList.CurPos := i;
          Break;
        end else
          TmpFileList[i].IsImage := False;
      end;
    finally
      TmpFileList.Free;
    end;
    if r <> 0 then
      r := ErrNotFoundInDirectory;
  end else
  begin
    r := ReadNewFile(FileName, False);
    if r = 0 then
      FileList.Clear;
  end;

  Result := (r = 0);
  // STDIN opens(NTn:ERROR_INVALID_HANDLE,9xn:ERROR_BAD_PATHNAME)
  // ̂Ƃ̓G[ɂȂ
  if ((r = ERROR_INVALID_HANDLE) or (r = ERROR_BAD_PATHNAME)) and (FileName = '') then
    Result := True;

  if not Result then
    LoadErrMsgBox(r, FileName);

  Invalidate;
end;

// LastFileInArc: TrueȂA[JCu̍Ō̃t@Cǂ
function TNkVMainForm.ReadNewFile(const FileName: string; LastFileInArc: Boolean): Integer;
var
  TmpFileList: TFileList;
  TmpTmpFileList: TFileList;
  oldCursor: HCURSOR;
  OldImageWidth, OldImageHeight: Integer;
  fd: TWin32FindData;
begin
  OldImageWidth := Image.Width;
  OldImageHeight := Image.Height;

  oldCursor := SetCursor(LoadCursor(0, IDC_WAIT));
  try
    // open O(ANZX^CXVO) Find Data 擾
    GetFindData(FileName, fd);
    Result := Image.LoadFromFile(FileName);
  finally
    SetCursor(oldCursor);
  end;

  case Result of
    0: // Image
      begin
        if ArcFileList.Archiver.Handle <> 0 then
        begin
          FreeLibrary(ArcFileList.Archiver.Handle);
          ArcFileList.Archiver.Handle := 0;
        end;
        Self.FileName := FileName;
        ArcFileName := '';
        FileCTime := fd.ftCreationTime;
        FileMTime := fd.ftLastWriteTime;
        FileATime := fd.ftLastAccessTime;
        UpdateViewState(OldImageWidth, OldImageHeight);
      end;
    nilfArchive:
      begin
        TmpFileList := TFileList.Create;
        try
          TmpFileList.Archiver.Handle := Image.hAMTmp;
          TmpFileList.LoadFromArchive(FileName);
          if LastFileInArc then
          begin
            TmpFileList.CurPos := TmpFileList.ItemCount;
            Result := ReadNextFileFromArchive(TmpFileList, True)
          end else
          begin
            TmpFileList.CurPos := -1;
            Result := ReadNextFileFromArchive(TmpFileList, False);
          end;
          if Result = 0 then
          begin
            if ArcFileList.Archiver.Handle <> 0 then
              FreeLibrary(ArcFileList.Archiver.Handle);
            TmpTmpFileList := ArcFileList;
            ArcFileList := TmpFileList;
            TmpFileList := TmpTmpFileList;
          end else
          begin
            // A[JCuɑΉt@C
            FreeLibrary(TmpFileList.Archiver.Handle);
            Result := ErrNotFoundInArchive;
          end;
        finally
          TmpFileList.Free;
        end;
      end;
  end;
end;

procedure TNkVMainForm.ReadNextFile(ForceNextFile: Boolean; PrevFlag: Boolean);
var
  dir, fname: string;
  r: Integer;
  st, ed, d, i: Integer;
begin
  if ArcFileList.Archiver.Handle = 0 then
  begin
    dir := DirName(FileName);
    fname := FileName;
  end else
  begin
    if not ForceNextFile then
    begin
      r := ReadNextFileFromArchive(ArcFileList, PrevFlag);
      if r = 0 then
        Exit;
    end;
    dir := DirName(ArcFileList.ArcFileName);
    fname := ArcFileList.ArcFileName;
  end;

  // FileListȂfBNg
  if FileList.ItemCount = 0 then
    FileList.LoadFromDirectory(dir, BaseName(fname));

  if PrevFlag then
  begin
    st := FileList.CurPos - 1;
    ed := 0;
    d := -1;
    if ed > st then Exit;
  end else
  begin
    st := FileList.CurPos + 1;
    ed := FileList.ItemCount - 1;
    d := 1;
    if ed < st then Exit;
  end;

  i := st;
  while True do
  begin
    if FileList[i].IsImage then
    begin
      if ReadNewFile(dir + '\' + FileList[i].Name, (not ForceNextFile and PrevFlag)) = 0 then
      begin
        FileList.CurPos := i;
        Break;
      end else
        FileList[i].IsImage := False;
    end;
    if i = ed then
      Break;
    Inc(i, d);
  end;
end;

procedure TNkVMainForm.ClientToImage(var pt: TPoint);
begin
  pt.x := (pt.x+Left) * 100 div Zoom;
  pt.y := (pt.y+Top) * 100 div Zoom;
end;

procedure TNkVMainForm.SetTitle;
var
  fn, info: string;
  dc: HDC;
  pt: TPoint;
  c: COLORREF;
  r, g, b: Integer;
begin
  with Image do
  begin
    if FileName = '' then
      fn := '<STDIN>'
    else if ArcFileName = '' then
      fn := BaseName(FileName)
    else
      fn := BaseName(ArcFileName) + '/' + FileName;

    if ShowColorValue then
    begin
      GetCursorPos(pt);
      pt := ScreenToClient(pt);
      dc := GetDC(Handle);
      c := GetPixel(dc, pt.x, pt.y);
      ReleaseDC(Handle, dc);
      ClientToImage(pt);

      r := c and $FF;
      g := c shr 8 and $FF;
      b := c shr 16 and $FF;

      info := SStr([' [R', IZStr(r, 3), ' G', IZStr(g, 3),' B', IZStr(b, 3), '] ',
        '(X', pt.x, ' Y', pt.y, ') ']);
    end else
      info := SStr([' (', Width, 'x', Height, 'x', BitPerPixelOrg, ') ']);

    if Stretch then
      info := info + 'Stretch Mode'
    else
      info := info + SStr([Zoom, '%']);

    Text := fn + info + ' - ' + AppName;
  end;
end;

procedure TNkVMainForm.PopupMenu(const pt: TPoint);
var
  i: Integer;
  Menu, SubMenu: HMENU;
  CmdMenu: TSCmdMenu;
begin
  CmdMenu := TSCmdMenu.Create(Cmd, MenuTable);
  if not Image.Loaded then
    CmdMenu.DisableItems([
      cmPrevImage,
      cmNextImage,
      cmPrevFile,
      cmNextFile,
      cmPicInfo,
      cmStretchMode,
      cmFullScreen,
      cmFitWindow,
      cmZoomIn,
      cmZoomOut,
      cmZoom50,
      cmZoom100,
      cmZoom150,
      cmZoom200,
      cmZoom300,
      cmZoom400,
      cmShowColorValue,
      cmCopyFileName,
      cmReloadImage,
      '{WallPaperMenu',
      cmSaveAs
    ]);
  if Image.Loaded and (FileName = '') then
    CmdMenu.DisableItems([
      cmPrevImage,
      cmNextImage,
      cmPrevFile,
      cmNextFile,
      cmCopyFileName,
      cmReloadImage
    ]);
  if FullScreen then
    CmdMenu.DisableItems([cmFitWindow]);
  if Image.FilterList.FileNames.Count = 0 then
    CmdMenu.DisableItems(['{PluginMenu']);

  CmdMenu.CheckItem(cmFullScreen, FullScreen);
  CmdMenu.CheckItem(cmStayOnTop, StayOnTop);
  CmdMenu.CheckItem(cmStretchMode, Stretch);
  CmdMenu.CheckItem(cmHideBorder, not Border);
  CmdMenu.CheckItem(cmShowColorValue, ShowColorValue);
  CmdMenu.CheckItem(cmScaleFilter, ScaleFilter);

  Menu := CmdMenu.MakeMenuHandle;
  CmdMenu.Free;

  SubMenu := GetSubMenu(Menu, GetMenuItemCount(Menu) - 5);
  with Image.FilterList do
    for i := 0 to FileNames.Count - 1 do
      AppendMenu(SubMenu, MF_STRING, IDM_PLUGINSETTING_FIRST+i, PChar(FileNames[i]));

  TrackPopupMenu(Menu, TPM_LEFTALIGN or TPM_LEFTBUTTON or TPM_RIGHTBUTTON,
    pt.x, pt.y, 0, Handle, nil);

  DestroyMenu(Menu);
end;

// windowTCYGɂ킹
procedure TNkVMainForm.FitWindow;
begin
  if Image.Loaded then
  begin
    if Maximized then
      Restore;
    ClientRect := Rect(0, 0, DispWidth, DispHeight);
  end;
end;

procedure TNkVMainForm.LoadErrMsgBox(Code: Integer; FileName: string);
var
  s: string;
begin
  if Code > 0 then
    s := '%1%n%n'+SysError(Code)
  else if Code = nilfDecodeErr then
    s := MsgDecodeError
  else if Code = ErrNotFoundInArchive then
    s := MsgNotFoundInArchive
  else if Code = ErrNotFoundInDirectory then
    s := MsgNotFoundInDirectory
  else
    s := MsgUnknownFormatError;
  if FileName = '' then
    FileName :=  '<STDIN>';
  ErrMsgBox(s, [FileName, SpiErrMsg[Image.DecodeResult], IStr(Image.DecodeResult), Image.ErrFilterFileName]);
end;

procedure TNkVMainForm.DrawGrid(dc: HDC; l, t, w, h: Integer; Bits: Pointer);
type
  PByteArray = ^TByteArray;
  TByteArray = array[0..MaxInt-1] of Byte;
var
  x, y, z: Integer;
  rc: TRect;
  br: HBRUSH;
begin
  if Zoom mod 100 <> 0 then
    Exit;
  z := Zoom div 100;
  if z < 4 then
    Exit;
  {
  for y := t to t+h-1 do
    if y mod z = 0 then
    begin
      MoveToEx(dc, 0-l, y-t, nil);
      LineTo(dc, w+l, y-t);
    end;
  for x := l to l+w-1 do
    if x mod z = 0 then
    begin
      MoveToEx(dc, x-l, 0-t, nil);
      LineTo(dc, x-l, h+t);
    end;
  }
  br := GetStockObject(BLACK_BRUSH);

  y := (t+z-1) div z * z;
  while y < (t+h) do
  begin
    rc.Left := -l;
    rc.Top := y-t;
    rc.Right := w+l+1;
    rc.Bottom := rc.Top+z+1;
    FrameRect(dc, rc, br);
    Inc(y, z*2);
  end;

  x := (l+z-1) div z * z;
  while x < (l+w) do
  begin
    rc.Left := x-l;
    rc.Top := -t;
    rc.Right := rc.Left+z+1;
    rc.Bottom := h+t+1;
    FrameRect(dc, rc, br);
    Inc(x, z*2);
  end;
end;

procedure TNkVMainForm.WMDropFiles(var Message: TWMDropFiles);
var
  Buf: array[0..MAX_PATH] of Char;
begin
  with Message do
  begin
    DragQueryFile(Drop, 0, Buf, SizeOf(Buf));
    DragFinish(Drop);
  end;
  LoadFile(Buf);
end;

procedure TNkVMainForm.WMPaint(var Message: TWMPaint);
var
  dc: HDC;
  ps: TPaintStruct;
  brush: HBrush;
  rc: TRect;
  r, b, fx, fy: Integer;
  DestBmi: TMaxBitmapInfo;
  DestBits: Pointer;
begin
  dc := BeginPaint(Handle, ps);
  rc := ClientRect;
  brush := GetStockObject(GRAY_BRUSH);

  // [hĂȂ΃O[ɓhĂ܂
  if not Image.Loaded then
  begin
    FillRect(dc, rc, brush);
    EndPaint(Handle, ps);
    Exit;
  end;

  if Stretch then
  begin
    with Image do
    begin
      if Width*rc.Bottom > Height*rc.Right then
      begin
        DispWidth := rc.Right;
        DispHeight := (Height * DispWidth) div Width;
      end else
      begin
        DispHeight := rc.Bottom;
        DispWidth := (Width * DispHeight) div Height;
      end;
    end;
    if DispHeight = 0 then
    begin
      EndPaint(Handle, ps);
      Exit;
    end;
  end;

  fx := 0;
  fy := 0;
  if FullScreen then
  begin
    fx := (GetSystemMetrics(SM_CXSCREEN)-DispWidth) div 2;
    if fx < 0 then fx := 0;
    fy := (GetSystemMetrics(SM_CYSCREEN)-DispHeight) div 2;
    if fy < 0 then fy := 0;
  end;
  SetViewportOrgEx(dc, fx, fy, nil);

  with Image do
  begin
    if rc.Right > DispWidth then
      r := DispWidth
    else
      r := rc.Right;
    if rc.Bottom > DispHeight then
      b := DispHeight
    else
      b := rc.Bottom;
    if (Zoom = 100) and not Stretch then
    begin
      SetDIBitsToDevice(dc, 0, 0, r, b, Left, Height-b, Top, Height, Bits,
        PBitmapInfo(@BitmapInfo)^, DIB_RGB_COLORS);
    end else
    begin
      ResizeDIB(PBitmapInfo(@DestBmi), DestBits, DispWidth, DispHeight,
        Left, Top, rc.Right, rc.Bottom, PBitmapInfo(@BitmapInfo)^, Bits, ScaleFilter);
      SetDIBitsToDevice(dc, 0, 0, r, b, 0, 0, 0, DestBmi.bmiHeader.biHeight, DestBits,
        PBitmapInfo(@DestBmi)^, DIB_RGB_COLORS);
      if ShowGrid then
        DrawGrid(dc, Left, Top, rc.Right, rc.Bottom, DestBits);
      GlobalFree(HGLOBAL(DestBits));
    end;
  end;

  with Image do
  begin
    // ܂ʂʂ
    if rc.right > DispWidth then
    begin
      FillRect(dc, Rect(rc.right, 0, DispWidth, DispHeight), brush);
      FillRect(dc, Rect(-fx, -fy, 0, DispHeight+fy+1), brush);
    end;
    if rc.bottom > DispHeight then
    begin
      FillRect(dc, Rect(0, rc.bottom, rc.Right, DispHeight), brush);
      FillRect(dc, Rect(0, -fy, rc.Right, 0), brush);
    end;
  end;

  EndPaint(Handle, ps);
end;

procedure TNkVMainForm.ShowAbout;
begin
  AboutBox(Handle, 'About '+AppName,
    AppName+' '+Version+#13#10+
    #13#10+
    Copyright+#13#10+
    #13#10+
    License+#13#10+
    #13#10+
    'Profile Folder: '+ProfDir);
end;

procedure TNkVMainForm.ShowPluginList;
var
  i: Integer;
  s, ss: string;
begin
  with Image.FilterList do
  begin
    s := SStr([FileNames.Count, ' plugin(s)'#13#10#13#10]);
    for i := 0 to FileNames.Count - 1 do
    begin
      if SpiFilter.Load(FileNames[i]) then
      begin
        ss := SpiFilter.Name;
        SpiFilter.Unload;
      end else
        ss := '(not found)';
      s := s + Formats[i] + #9 + FileNames[i] + ': ' + ss + #13#10;
    end;
  end;
  MsgBox(Handle, 'Plug-in List', s);
end;

procedure TNkVMainForm.ShowPluginConfig(const PluginFileName: string);
begin
  with Image.FilterList do
  begin
    if SpiFilter.Load(PluginFileName) then
    begin
      if SpiFilter.ConfigDlg(Handle) = -1 then
        FormatMsgBox(Handle, AppName, MsgPluginConfDlgNotImplement, [PluginFileName], MB_ICONINFORMATION);
      SpiFilter.UnLoad;
    end else
      ErrMsgBox(MsgPluginLoadError, [PluginFileName])
  end;
end;

procedure TNkVMainForm.ZoomImage(Value: Integer);
var
  rc: TRect;
begin
  if (Value > 0) and (not Stretch) then
  begin
    with Image do
    begin
      Left := Left * 100 div Zoom;
      Top  := Top * 100 div Zoom;
      Zoom := Value;
      DispWidth := Width * Zoom div 100;
      DispHeight := Height * Zoom div 100;
      Left := Left * Zoom div 100;
      Top  := Top * Zoom div 100;
      // ʒu
      rc := ClientRect;
      if DispWidth > rc.Right then
      begin
       if Left+rc.Right > DispWidth then Left := DispWidth - rc.Right;
      end else
        Left := 0;
      if DispHeight > rc.Bottom then
      begin
        if Top+rc.Bottom > DispHeight then Top := DispHeight - rc.Bottom;
      end else
         Top := 0;
    end;
  end;
  SetTitle;
  InvalidateRect(Handle, nil, False);
end;

procedure TNkVMainForm.ZoomInOut(OutFlg: Boolean);
var
  i, z: Integer;
begin
  z := Zoom;
  i := 0;
  while  i <= High(ZoomTable) do
  begin
    if OutFlg then
    begin
      if ZoomTable[i] >= Zoom then
        Break;
      z := ZoomTable[i];
    end
    else begin
      z := ZoomTable[i];
      if ZoomTable[i] > Zoom then
        Break;
    end;
    Inc(i);
  end;
  ZoomImage(z);
end;

// x, y: ړ
procedure TNkVMainForm.ScrollImage(x, y: Integer);
var
  rc: TRect;
  oldleft, oldtop: Integer;
//  dc: HDC;  rgn: HRGN;   rrc:TRECT;
begin
  with Image do
  begin
    if not Loaded then
      Exit;

    rc := ClientRect;
    oldleft := Left;
    oldtop := Top;

    if DispWidth > rc.Right then
    begin
      Inc(Left, x);
      if Left < 0 then
        Left := 0;
      if Left > DispWidth - rc.Right then
         Left := DispWidth - rc.Right;
    end;
    if DispHeight > rc.Bottom then
    begin
      Inc(Top, y);
      if Top < 0 then
        Top := 0;
      if Top > DispHeight - rc.Bottom then
        Top := DispHeight - rc.Bottom;
    end;
    if ScrollWindow(Handle, oldleft-Left, oldtop-Top, nil, @rc) then
      Update;

    //dc := GetDC(hWnd);
    //ScrollDC(dc, (oldleft-Left), (oldtop-Top), rc, rc, 0, @rrc);
    //ReleaseDC(hWnd, dc);
    //InvalidateRect(hWnd, @rrc, False);
  end;
end;

procedure TNkVMainForm.CopyFileNameToClipBoard;
var
  s: string;
  h: HGLOBAL;
  p: PChar;
begin
  if ArcFileName = '' then
    s := FileName
  else
    s := ArcFilename;
  h := GlobalAlloc(GHND, Length(s)+1);
  p := GlobalLock(h);
  Move(s[1], p^, Length(s));
  GlobalUnlock(h);
  if OpenClipBoard(Handle) then
  begin
    EmptyClipBoard;
    SetClipBoardData(CF_TEXT, h);
    CloseClipBoard;
  end else
    GlobalFree(h);
end;

procedure TNkVMainForm.DoCmd(const CmdName: string);
var
  i, r: Integer;
begin
  if CmdName = '' then
    Exit
  else if CmdName = cmStayOnTop then
    StayOnTop := not StayOnTop
  else if CmdName = cmHideBorder then
  begin
    if not FullScreen then
      Border := not Border;
    if Border then
      NormalCursor := LoadCursor(HInstance, 'HANDOPEN')
    else
      NormalCursor := LoadCursor(0, IDC_ARROW);
    SetCursor(NormalCursor);
  end
  else if CmdName = cmScaleFilter then
  begin
    ScaleFilter := not ScaleFilter;
    Invalidate;
  end
  else if CmdName = cmShowGrid then
  begin
    ShowGrid := not ShowGrid;
    Invalidate;
  end
  else if CmdName = cmNextWindow then
    ActivateNextForm(False)
  else if CmdName = cmPrevWindow then
    ActivateNextForm(True)
  else if CmdName = cmPluginList then
    ShowPluginList
  else if CmdName = cmAbout then
    ShowAbout
  else if CmdName = cmExit then
    Close;

  if not Image.Loaded then
    Exit;

  // Image Loaded
  if CmdName = cmPicInfo then
  begin
    if not (IsWindow(InfoDlg) and IsWindowVisible(InfoDlg)) then
    begin
      PicInfoDlg.ShowModeless;
      InfoDlg := PicInfoDlg.Handle;
    end;
  end
  else if CmdName = cmFitWindow then
  begin
    if not FullScreen then
      FitWindow;
  end
  else if CmdName = cmZoomIn then
  begin
    Stretch := False;
    ZoomInOut(False);
  end
  else if CmdName = cmZoomOut then
  begin
    Stretch := False;
    ZoomInOut(True);
  end
  else if Copy(CmdName, 1, 4) = Copy(cmZoom100, 1, 4) then
  begin
    Stretch := False;
    ZoomImage(SInt(Copy(CmdName, 5, Length(CmdName)-5+1)));
  end
  else if CmdName = cmFullScreen then
    FullScreen := not FullScreen
  else if CmdName = cmSetWallPaper then
    SetWallPaper(False)
  else if CmdName = cmSetWallPaperTile then
    SetWallPaper(True)
  else if CmdName = cmStretchMode then
  begin
    Stretch := not Stretch;
    Left := 0;
    Top := 0;
    ZoomImage(Zoom);
  end
  else if CmdName = cmSmoothZoomIn then
  begin
    for i := 0 to 4 do
    begin
      ZoomImage(Zoom+1);
      Update;
    end;
  end
  else if CmdName = cmSmoothZoomOut then
  begin
    for i := 0 to 4 do
    begin
      ZoomImage(Zoom-1);
      Update;
    end;
  end
  else if CmdName = cmScrollLeft then
    ScrollImage(-Scroll,0)
  else if CmdName = cmScrollRight then
    ScrollImage(Scroll, 0)
  else if CmdName = cmScrollUp then
    ScrollImage(0,-Scroll)
  else if CmdName = cmScrollDown then
    ScrollImage(0, Scroll)
  else if CmdName = cmScrollTop then
    ScrollImage(0, -DispHeight)
  else if CmdName = cmScrollBottom then
    ScrollImage(0, DispHeight)
  else if CmdName = cmScrollTopBottom then
  begin
    if Top = DispHeight - ClientRect.Bottom then
      ScrollImage(0, -DispHeight)
    else
      ScrollImage(0, DispHeight);
  end
  else if CmdName = cmShowColorValue then
  begin
    ShowColorValue := not ShowColorValue;
    if ShowColorValue then
      NormalCursor := LoadCursor(0, IDC_CROSS)
    else if Border then
      NormalCursor := LoadCursor(HInstance, 'HANDOPEN')
    else
      NormalCursor := LoadCursor(0, IDC_ARROW);
    SetCursor(NormalCursor);
    SetTitle;
  end
  else if CmdName = cmSaveAs then
    SaveAs;

  if FileName = '' then
    Exit;

  // Image Loaded and not <STDIN>
  if CmdName = cmNextImage then
    ReadNextFile(False, False)
  else if CmdName = cmPrevImage then
    ReadNextFile(False, True)
  else if CmdName = cmNextFile then
    ReadNextFile(True, False)
  else if CmdName = cmPrevFile then
    ReadNextFile(True, True)
  else if CmdName = cmRefreshFileList then
    FileList.Clear
  else if CmdName = cmCopyFileName then
    CopyFileNameToClipBoard
  else if CmdName = cmReloadImage then
  begin
    if ArcFileName = '' then
      r := ReadNewFile(FileName, False)
    else
      r := ReadNewFile(ArcFileName, False);
    Invalidate;
    if r <> 0 then
      LoadErrMsgBox(r, FileName);
  end;
  {
  IDM_ARCHIVEFILELIST:
    begin
      for i := 0 to ArcFileList.ItemCount - 1 do
        s := s + PChar(ArcFileList[i]) + #13#10;
      MessageBox(Handle, PChar(s), AppName + '', 0);
    end;
  }
end;

procedure TNkVMainForm.WMMouseWheel(var Message: TMessage);
begin
  with Message do
    if (GetKeyState(VK_LBUTTON) < 0) and not Stretch then
    begin
      if WParam > 0 then
        DoCmd(cmSmoothZoomIn)
      else
        DoCmd(cmSmoothZoomOut);
    end else
    begin
      if WParam > 0 then
        DoCmd(cmPrevImage)
      else
        DoCmd(cmNextImage);
    end;
end;

procedure TNkVMainForm.WMCommand(var Message: TWMCommand);
begin
  with Message do
    case ItemID of
      IDM_PLUGINSETTING_FIRST .. IDM_PLUGINSETTING_LAST:
        ShowPluginConfig(Image.FilterList.FileNames[ItemID - IDM_PLUGINSETTING_FIRST]);
    else
      DoCmd(Cmd.Name(ItemID));
    end;
end;

procedure TNkVMainForm.WMContextMenu(var Message: TMessage);
var
  pt: TPoint;
begin
  if GetContextMenuPos(Handle, Message.LParam, 0, 0, pt) then
    PopupMenu(pt)
  else
    inherited;
end;

procedure TNkVMainForm.WMSetCursor(var Message: TWMSetCursor);
begin
  with Message do
    if HitTest = HTCLIENT then
    begin
      SetCursor(NormalCursor);
      Result := 1;
    end;
  inherited;
end;

procedure TNkVMainForm.WMLButtonDown(var Message: TWMLButtonDown);
var
  pt: TPoint;
begin
  if Border or FullScreen then
  begin
    SetCapture(Handle);
    SetCursor(GrabCursor);
    GetCursorPos(pt);
    oldpt := pt;
  end else
  begin
    ReleaseCapture;
    //SetCursor(LoadCursor(0, IDC_ARROW));
    Perform(WM_NCLBUTTONDOWN, HTCAPTION, 0);
  end;
end;

procedure TNkVMainForm.WMLButtonUp(var Message: TWMLButtonUp);
begin
  ReleaseCapture;
  SetCursor(NormalCursor);
end;

procedure TNkVMainForm.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
  DoCmd(cmFitWindow);
end;

procedure TNkVMainForm.WMMouseMove(var Message: TWMMouseMove);
var
  pt: TPoint;
begin
  with Message do
    // StretchŃXN[悤ƂĂcaption_uNbN
    // ő剻悤ƂƗ̂Stretch ModêƂ͉Ȃ
    if Image.Loaded and (Keys = MK_LBUTTON) and not Stretch then
    begin
      GetCursorPos(pt); // ƂȂ window ÔƂWƂȂ
      ScrollImage(oldpt.x - pt.x, oldpt.y - pt.y);
      oldpt := pt;
    end
    else if ShowColorValue then
      SetTitle;
end;

procedure TNkVMainForm.WMSize(var Message: TWMSize);
begin
  with Message do
    if Image.Loaded and not Stretch then
    begin
      if DispWidth > Width then
      begin
        if Left+Width > DispWidth then
          Left := DispWidth - Width;
      end else
        Left := 0;
      if DispHeight > Height then
      begin
        if Top+Height > DispHeight then
          Top := DispHeight - Height;
      end else
        Top := 0;
    end;
end;

procedure LoadConfig;
var
  Name, Arg: string;
  ArgCount: Integer;
begin
  with TSConfFile.Create(ProfDir + ConfigFile) do
    try
      while GetLine(Name, ArgCount, Arg) do
        if Name = 'KeepViewPosition' then
          KeepViewPositon := StrToBool(Arg);
    finally
      Free;
    end;
end;

var
  r: Integer;
begin
  ProfDir := GetProfDir('nkv') + '\';
  LoadConfig;

  Cmd := TSCmd.Create(CmdList, DefaultKeys, ProfDir + KeyMapFile);
  FileList := TFileList.Create;
  ArcFileList := TFileList.Create;
  Image := TNkVImage.Create;
  MainForm := TNkVMainForm.CreateDefaultPos(AppName, 320, 240);
  try
    if MainForm.Handle = 0 then
      Halt(255);

    SetMainForm(MainForm);

    r := LoadPluginListFile(ProfDir + PluginListFile);
    if r <> 0 then
      MainForm.ErrMsgBox('%1%n%n'+SysError(r), [ProfDir + PluginListFile]);

    if not MainForm.LoadFile(GetLongFileName(ParamStr(1))) then
      Exit;

    //MainForm.DoCmd(cmZoom50);
    //MainForm.DoCmd(cmScaleFilter);
    PicInfoDlg := TPicInfoDlg.CreateByResource(MainForm.Handle, 'INFODIALOG');

    MainForm.Show;

    ExitCode := MessageLoopWithAccelAndDialog(MainForm.Handle, Cmd.AccHandle, InfoDlg);

    PicInfoDlg.Free;
  finally
    MainForm.Free;
    Image.Free;
    FileList.Free;
    ArcFileList.Free;
    Cmd.Free;
  end;
end.

